--[[ LibMerisioux

     Generic addon for exchanging RP data.

     {
       prefix = "",
       override = "",
       suffix = "",
       title = "",
       cgender = "",
       pgender = "",
       description = "",
       biography = "",
       meta = "",
       flags = "",
       version = "",
       age = "",
       height = "",
       weight = "",
       currently = "",
       custom_name = "",
       custom_data = "",
     }

     snoop(char, callback, query)
     	=> calls callback(char, data, error)
	If query is true, always query.  If query is false, never query.
	Otherwise, tries to use cached data if it's under a minute old,
	and uses cached data if a query fails.  WARNING:  If you want to
	mess with the table, copy it!  I am not copying tables.

     flaunt(table)
     flaunt(field, value)
     	=> Sets data about this character.  If the argument is a table,
	it's treated as { field = value } pairs.  A boolean false clears
	a value.

     setflag(c, state, flag_set)
        => Sets the given flag (or unsets it if state is false).
	Note:  For purposes of this function, 'nil' is a kind of true;
	you must use an explicit false to unset a flag. Returns
	a string of the flags removed, and the resulting flag set.

	The old behavior (maintaining a separate editing_flags state)
	has been judged horrific and removed. Yes, that's a breaking
	change. Sorry.

     addon(addon_identifier, newer)
     	=> Tells LibMerisioux that it is being used by 'addon_identifier',
	and provides a callback to use if a newer version of that particular
	addon is observed.

     info()
        => displays info about character

     list(flags)
        => a list of characters currently known (who have all of the
	   listed flags)

     fields()
        => a list of fields currently known.
		field = {
		  max = n,
		  help = "description",
		}
	Again, NOT A COPY.  Copy it before you change it.

     flags()
        => table, string
	Table is a list of known flags, string is a canonical sorted order
	for the flags.
	Again, NOT A COPY.  Do not change this.
		character => { name =, description =, excludes = }

     prettyname(name, data)
        => A prettied up string from the prefix, override, and suffix fields.

]]--

local addoninfo, meri = ...
Library = Library or {}
Library.LibMerisioux = meri
local rar = Library.LibRarian
meri.data_version = "0"
meri.version = "0.17-151128-15:56:25"
meri.callbacks = {}
meri.callback_times = {}

local slower = string.lower

local function tcopy(t)
  local u = {}
  for k, v in pairs(t) do
    u[k] = v
  end
  return u
end

local function tkeys(t)
  local u = {}
  for k, v in pairs(t) do
    u[#u + 1] = k
  end
  return u
end

local printf = Library.printf.printf
local sprintf = Library.printf.sprintf

meri.debugging = false
meri.lfrp_allow_self = false

function meri.debug(...)
  if meri.debugging then
    printf(...)
  end
end

-- logical groupings: cfo lhw enm vp ri PdR
meri.sorted_flags = "cfolhwenmvpriPdR"

meri.flagtable = {
  e = { name = 'ERP',
	description = 'ERP-friendly.',
	excludes = 'mn', },
  n = { name = 'No ERP',
	description = 'No ERP.',
	excludes = 'em', },
  m = { name = 'Mature',
        description = 'Sexual themes acceptable when they fit the story.',
	excludes = 'en', },
  v = { name = 'Conflict',
        description = 'Comfortable with high-conflict storylines and RP.',
	excludes = 'p', },
  p = { name = 'Peaceful',
        description = 'Uncomfortable with high-conflict storylines and RP.',
	excludes = 'v', },
  h = { name = 'Helper',
        description = 'Newbie-helper',
	excludes = 'w' },
  w = { name = 'Newbie',
        description = 'New to the game or to RP.',
	excludes = 'h', },
  l = { name = 'Looking',
        description = 'Looking for contacts or RP.',
	excludes = '', },
  f = { name = 'Full-time',
        description = 'Full-time RP.',
	excludes = 'o', },
  c = { name = 'IC',
  	description = 'In character.',
	excludes = 'o', },
  o = { name = 'OOC',
  	description = 'Out of character.',
	excludes = 'cf', },
  r = { name = 'Lore',
        description = "Following Rift's lore/setting.",
	excludes = 'i', },
  i = { name = 'Non-lore',
        description = "Ignoring Rift's lore/setting.",
	excludes = 'r', },
  P = { name = 'PvP',
        description = 'PvP to resolve fights.',
	excludes = '', },
  d = { name = 'Dice',
        description = 'Dice to resolve fights.',
	excludes = '', },
  R = { name = 'Roleplay',
        description = 'Roleplay to resolve fights.',
	excludes = '', },
}

function meri.flags_from(s)
  local t = ""
  for i = 1, #s do
    local c = s:sub(i, i)
    if meri.flagtable[c] then
      t = t .. c
    end
  end
  return t
end

function meri.addon(identifier, callback)
  local addoninfo = Inspect.Addon.Detail(identifier)
  if addoninfo then
    meri.addon_name = identifier
    meri.addon_info = string.format("%s:%s", identifier, addoninfo.toc.Version or 'unknown')
    -- this used to try to update the stored data with the new version.
    -- that could interfere with the check for saved data; instead, we just
    -- figure you'll eventually save things with the new version, and
    -- then things will work. Note that "eventually" may be "as soon as the
    -- stored data callback completes".
    meri.addon_callback = callback
  else
    printf("Help:  LibMerisioux was told to use addon ID '%s', but can't find it.", tostring(identifier))
  end
end

function meri.fields()
  return meri.template
end

function meri.flags()
  return meri.flagtable, meri.sorted_flags
end

meri.template = {
  prefix = { max = 15,
    help = "A prefix to display before the character's name.",
  },
  override = { max = 15,
    help = "A name to display in place of the character name.",
  },
  suffix = { max = 15,
    help = "A suffix to display after the character's name.",
  },
  title = { max = 35,
    help = "A title to display.",
  },
  cgender = { max = 15,
    help = "Character gender."
  },
  pgender = { max = 15,
    help = "Player gender."
  },
  description = { max = 2048,
    help = "A description of externally-observable traits or details.",
  },
  biography = { max = 2048,
    help = "Character backstory, need not be externally-observable.",
  },
  meta = { max = 2048,
    help = "OOC notes about preferred RP style or play; notes for players, not characters.",
  },
  flags = { max = 26,
    help = "A series of flags indicating things about your character.",
  },
  stamp = { max = 16,
    help = "A timestamp for this bio."
  },
  age = { max = 32,
    help = "The character's approximate age."
  },
  height = { max = 32,
    help = "The character's height."
  },
  weight = { max = 32,
    help = "The character's weight."
  },
  currently = { max = 64,
    help = "Current activity/status."
  },
  custom_name = { max = 24,
    help = "Custom field name."
  },
  custom_data = { max = 80,
    help = "Custom field data."
  },
}

-- just in case...
function meri.sanity_check()
  local missed = {}
  local extra = {}
  for k, v in pairs(meri.flagtable) do
    missed[k] = true
  end
  for c in string.gmatch(meri.sorted_flags, '.') do
    if missed[c] then
      missed[c] = nil
    else
      extra[c] = true
    end
  end
  for k, v in pairs(extra) do
    printf("Warning:  Sort order specified for nonexistent flag %s.", k)
  end
  for k, v in pairs(missed) do
    printf("Warning:  No sort order specified for flag %s.", k)
  end
  if meri.self then
    if meri.self.version then
      if meri.self.version ~= meri.data_version then
        printf("Whoops, someone forgot to add version upgrades to LibMerisioux.")
	printf("This profile is version %d, LibMerisioux is version %d.", meri.self.version, meri.data_version)
      end
    else
      -- probably harmless
      meri.self.version = meri.data_version
    end
  end
end

function meri.variables_loaded(event, name)
  if name == 'LibMerisioux' then
    LibMerisiouxData = LibMerisiouxData or { shards = {} }
    local shard = Inspect.Shard().name
    LibMerisiouxData.shards[shard] = LibMerisiouxData.shards[shard] or {}
    if not LibMerisiouxData.last_version_seen then
      LibMerisiouxData.last_version_seen = meri.version
    else
      local comp = Library.LibVersion.compare(LibMerisiouxData.last_version_seen, meri.version)
      if comp == -1 then
        LibMerisiouxData.last_version_seen = meri.version
      elseif comp == 1 then
        printf("LibMerisioux:  Saw newer version %s, you have %s.",
	    LibMerisiouxData.last_version_seen, meri.version)
      end
    end
    meri.data = LibMerisiouxData.shards[shard]
    meri.whoami = slower(Inspect.Unit.Detail('player').name)
    meri.data[meri.whoami] = meri.data[meri.whoami] or { }
    meri.self_metadata = meri.data[meri.whoami]
    if not meri.self_metadata.data then
      meri.self_metadata.data = { version = meri.data_version }
    end
    -- add lfrp data
    if not meri.self_metadata.lfrp_data then
      meri.self_metadata.lfrp_data = {}
    end
    -- add lfrp config
    if not meri.self_metadata.lfrp_config then
      meri.self_metadata.lfrp_config = { channels = { }, active = false }
    end

    meri.self = meri.self_metadata.data
    -- Most likely, you wrote it in 1970.
    if not meri.self.stamp then
      printf("Found no existing stamp. Adding.")
      meri.self.stamp = 1
    end
    -- remember values which were stored locally, never delete them
    meri.self_metadata.self = true
    -- this will call us back when it knows what was on the server, if
    -- anything.
    rar.retrieve(meri.whoami, 'merisioux_data', meri.maybe_newer_data, true)
    meri.sanity_check()
  end
end

function meri.clean_variables()
  if not meri.data then
    return
  end
  for k, v in pairs(meri.data) do
    if v.error and not v.self and not v.data then
      meri.data[k] = nil
    end
  end
end

function meri.inflater(x)
  local inflate = zlib.inflate()
  return inflate(x)
end

function meri.parse(data, error)
  if not data then
    return data, error
  end
  local status, value
  -- Future directions: In a future release, we will stop doing the
  -- compression in LibMerisioux, because LibRarian does it automatically.
  -- So for now, we will accept either of two results.
  if type(data) == 'table' then
    value = data
  else
    status, value = pcall(meri.inflater, data)
    if status then
      local func, err = loadstring('return ' .. value)
      if func then
	setfenv(func, {})
	status, value = pcall(func)
	if not status then
	  return nil, "Failed to execute: " .. value
	else
	  if type(value) ~= 'table' then
	    return nil, "Invalid data format recieved."
	  end
	end
      else
	return nil, "Couldn't parse returned value."
      end
    end
  end

  local found_version = nil
  local cleaned = {}
  for k, v in pairs(value) do
    if meri.template[k] then
      if type(v) == 'string' then
        cleaned[k] = string.sub(v, 1, meri.template[k].max)
      elseif type(v) == 'number' then
        cleaned[k] = v
      end
    elseif k == 'libversion' then
      -- ignore "dev" versions.
      if not v:match('dev') then
        local comp = Library.LibVersion.compare(LibMerisiouxData.last_version_seen, v)
        if comp == -1 then
	  printf("LibMerisioux:  Saw newer version %s, you have %s.",
	      v, meri.version)
	  LibMerisiouxData.last_version_seen = v
        end
      end
    elseif k == 'addonversion' then
      -- The addon we're using might care, if we have one.
      local name, version = string.match(v, '([^:]*):(.*)')

      if name == meri.addon_name and meri.addon_callback then
	Utility.Dispatch(function() meri.addon_callback(version) end, meri.addon_name, 'new version callback')
      end
    elseif k == 'version' and type(v) == 'number' then
      -- we might someday use this to translate from older
      -- versions...
      found_version = v
      if v > meri.data_version then
	printf("Warning:  LibMerisioux version %d got a version %d profile.  You may need to upgrade.", meri.data_version, v)
      end
    end
  end
  return cleaned, nil
end

function meri.maybe_newer_data(char, identifier, data, error)
  if char ~= meri.whoami or identifier ~= 'merisioux_data' then
    meri.update_local(true)
    return
  end

  if error then
    -- nope, no newer data
    meri.update_local(true)
    return
  end

  data, error = meri.parse(data, error)
  if not data then
    meri.update_local(true)
    return
  end

  local newer = nil
  local older = nil
  if data.stamp then
    if data.stamp > meri.self.stamp then
      newer = true
    elseif data.stamp < meri.self.stamp then 
      older = true
    end
  elseif meri.self.stamp == 1 then
    -- the data on the server had no timestamp, so it's from 0.17. You
    -- haven't updated since you got this newer version, and anyway, if you
    -- had updated, it would have gone to the server, right? So.
    newer = true
    -- we don't have a stamp from the server, so we'll stamp it as "now".
    data.stamp = Inspect.Time.Server()
  else
    -- Your data is definitely newer than the server's data.
    older = true
  end
  -- if there's a mismatch, save the older data somewhere else. Also back up
  -- lfrp settings.
  if older then
    printf("Found existing data on server, but it's older. Saving it as %s_server.", meri.whoami)
    meri.data[meri.whoami .. '_server'] = { stamp = data.stamp or Inspect.Time.Server(), data = data, self = true, lfrp_data = tcopy(meri.self_metadata.lfrp_data), lfrp_config = tcopy(meri.self_metadata.lfrp_config) }
  elseif newer then
    printf("Found existing data on server, and it's newer. Saving any local settings as %s_local.", meri.whoami)
    meri.data[meri.whoami .. '_local'] = { stamp = meri.self.stamp or 1, data = meri.self, self = true, lfrp_data = tcopy(meri.self_metadata.lfrp_data), lfrp_config = tcopy(meri.self_metadata.lfrp_config) }
    meri.data[meri.whoami] = { stamp = data.stamp or Inspect.Time.Server(), data = data, self = true, lfrp_data = tcopy(meri.self_metadata.lfrp_data), lfrp_config = tcopy(meri.self_metadata.lfrp_config) }
    meri.self_metadata = meri.data[meri.whoami]
    meri.self = meri.self_metadata.data
  end
  meri.update_local(true)
end

function meri.snoop_callback(char, identifier, data, error)
  if not meri.callbacks[char] then
    return
  end
  -- and remove the callback time record for this character, so we don't
  -- report it as timed out later.
  meri.callback_times[char] = nil
  data, error = meri.parse(data, error)
  if char ~= meri.whoami then
    if data then
      meri.data[char] = { stamp = Inspect.Time.Server(), data = data }
    else
      -- use old value if we have it
      if meri.data[char] and meri.data[char].data then
	data = meri.data[char].data
	error = meri.data[char].error or error
      else
	meri.data[char] = { stamp = Inspect.Time.Server(), error = error }
      end
    end
  end
  for _, cb in ipairs(meri.callbacks[char]) do
    -- cb[1] = callback, cb[2] = addon identifier
    Utility.Dispatch(function() cb[1](char, data, error) end, cb[2], 'LibMerisioux data callback')

  end
  meri.callbacks[char] = nil
end

-- query can be true, false, or nil. nil means use an existing value
-- if it's recent. false means use only local cache. true means always
-- query even if there's an existing cached value.
function meri.snoop(char, callback, query)
  local addon = Inspect.Addon.Current()
  meri.debug("Attempting snoop for %s [query %s].", char, tostring(query))
  if type(char) ~= 'string' then
    Utility.Dispatch(function() callback(char, nil, "Character name must be a string.") end, addon, 'LibMerisioux snoop callback')
    return
  end
  char = string.lower(char)
  meri.debug("Lowercased char.")
  if char == meri.whoami then
    meri.debug("It's me.")
    if meri.self then
      Utility.Dispatch(function() callback(meri.whoami, meri.self, nil) end, addon, 'LibMerisioux snoop callback')
    else
      Utility.Dispatch(function() callback(meri.whoami, nil, 'ENOENT') end, addon, 'LibMerisioux snoop callback')
    end
    return
  end
  if not rar then
    meri.debug("No rar.")
    Utility.Dispatch(function() callback(char, nil, "Can't find LibRarian to snoop on people") end, addon, 'LibMerisioux snoop callback')
    return
  end
  meri.debug("Okay, now what.")
  if query == false then
    meri.debug("No query.")
    if meri.char.data then
      meri.debug("No query, using existing data.")
      Utility.Dispatch(function() callback(char, meri.data[char].data, meri.data[char].error) end, addon, 'LibMerisioux snoop callback')
      return
    else
      meri.debug("No query, also no data.")
      Utility.Dispatch(function() callback(char, nil, "ENOENT") end, addon, 'LibMerisioux snoop callback')
      return
    end
  end
  if meri.data[char] and not query then
    local limit = meri.debugging and 10 or 600
    if Inspect.Time.Server() - (meri.data[char].stamp or 0) < limit then
      meri.debug("Using cached result for %s, because old data is %d/%d seconds old.", char, (Inspect.Time.Server() - (meri.data[char].stamp or 0)), limit)
      Utility.Dispatch(function() callback(char, meri.data[char].data, meri.data[char].error) end, addon, 'LibMerisioux snoop callback')
      return
    end
  end
  if not meri.callbacks[char] then
    meri.debug("Setting up new callback for %s.", char)
    meri.callbacks[char] = { { callback, Inspect.Addon.Current() } }
    meri.callback_times[char] = Inspect.Time.Frame()
    rar.retrieve(char, 'merisioux_data', meri.snoop_callback)
  else
    meri.debug("Adding callback for %s.", char)
    meri.callbacks[char][#meri.callbacks[char] + 1] = { callback, Inspect.Addon.Current() }
  end
  meri.debug("Got here.")
end

function meri.update_hook()
  local now = Inspect.Time.Frame()
  local removes = {}
  for char, time in pairs(meri.callback_times) do
    if now - time >= 4 then
      meri.snoop_callback(char, identifier, nil, 'Timed out.')
      table.insert(removes, char)
    end
  end
  for _, char in ipairs(removes) do
    meri.callback_times[char] = nil
  end
  -- expire old handles
  for k, v in pairs(meri.lfrp_handles) do
    if v.expire < now then
      if v.callback then
        meri.debug("Closing out expired handle %s.", k)
        v.callback(k, true, v.responses)
      end
      meri.lfrp_handles[k] = nil
    end
  end
end

function meri.flaunt_specific(field, value)
  if value == false then
    if meri.template[field] then
      meri.self[field] = nil
      return
    else
      printf("I've never heard of a '%s', how can I not flaunt it?",
	tostring(field))
    end
  end
  if type(value) ~= 'string' then
    printf("Oh, I can't flaunt *that*, it's a %s.", type(value))
    return
  end
  if meri.template[field] then
    meri.self[field] = string.sub(value, 1, meri.template[field].max)
  else
    printf("I've never heard of a '%s', how can I flaunt it?",
      tostring(field))
  end
end

function meri.update_local_callback(extra_callback, failure)
  if failure then
    printf("Warning: An error occurred trying to save your profile on the server. It may be too large.")
    printf("Your profile should still work, but may not be recoverable if your local data files get lost.")
    if type(failure) == 'string' then
      if #failure > 60 then
        failure = failure:sub(1, 60) .. "..."
      end
      printf("The error was: %s", failure)
    end
    if extra_callback then
      extra_callback(failure)
    end
  end
end

function meri.update_local(skip_stamp, extra_callback)
  if meri.self then
    meri.self.version = meri.data_version
    if not skip_stamp then
      meri.self.stamp = Inspect.Time.Server()
    end
    if meri.addon_info then
      meri.self.addonversion = meri.addon_info
    else
      meri.self.addonversion = nil
    end
    meri.self.libversion = meri.version
    local verbose = Utility.Serialize.Inline(meri.self)
    local deflate = zlib.deflate(zlib.BEST_COMPRESSION)
    local deflated, eof, bytes_in, bytes_out = deflate(verbose, "finish")
    local error = rar.store('merisioux_data', 'public', deflated, true, function(...) meri.update_local_callback(extra_callback, ...) end)
    return bytes_in, bytes_out, error
  else
    rar.store('merisioux_data', 'public', nil, true, function(...) meri.update_local_callback(extra_callback, ...) end)
    return 0, 0
  end
end

function meri.setflag(c, state, flag_set)
  if not meri.flagtable[c] then
    return '', flag_set
  end
  if not flag_set then
    return '', ''
  end
  if state ~= false then
    local removed = ''
    if not flag_set:find(c) then
      flag_set = flag_set .. c
    end
    if meri.flagtable[c].excludes and meri.flagtable[c].excludes ~= '' then
      for c2 in string.gmatch(meri.flagtable[c].excludes, '.') do
        if string.find(flag_set, c2) then
	  flag_set = flag_set:gsub(c2, '')
          removed = removed .. c2
        end
      end
    end
    return removed, flag_set
  else
    flag_set = flag_set:gsub(c, '')
    return '', flag_set
  end
end

function meri.flaunt(field_or_table, value, extra_callback)
  if not field_or_table then
    return nil, "You can't flaunt it if you haven't got it."
  end
  if value then
    meri.flaunt_specific(field_or_table, value)
    return meri.update_local()
  end
  if type(field_or_table) ~= 'table' then
    return nil, string.format("I really don't see how I can flaunt a %s.", type(field_or_table))
  end
  for k, v in pairs(field_or_table) do
    meri.flaunt_specific(k, v)
  end
  return meri.update_local(false, extra_callback)
end

function meri.prettyname(name, data)
  local prefix = data and data.prefix or ''
  local suffix = data and data.suffix or ''
  local override = data and data.override or ''

  if override ~= '' then
    name = override
  end
  name = string.upper(string.sub(name, 1, 1)) .. string.lower(string.sub(name, 2))

  local pretty = prefix

  if prefix ~= '' and name ~= '' then
    pretty = pretty .. ' '
  end

  pretty = pretty .. name

  if name ~= '' and suffix ~= '' then
    pretty = pretty .. ' '
  end

  pretty = pretty .. suffix

  return pretty
end

function meri.list(flagstring)
  local flags = {}
  if flagstring then
    for i = 1, #flagstring do
      flags[#flags + 1] = flagstring:sub(i, i)
    end
  end
  local known = {}
  for char, details in pairs(meri.data) do
    local stash = true
    if #flags > 0 then
      if details.data and details.data.flags then
        for i = 1, #flags do
	  if not details.data.flags:match(flags[i]) then
	    stash = false
	    break
	  end
	end
      else
        stash = false
      end
    end
    if stash then
      known[#known + 1] = char
    end
  end
  return known
end

-- outstanding handles which need processing:
meri.lfrp_handles = {
  -- handle = { expire = stamp, callback = callback }
}

-- limit spam
meri.seen_lfrp_handles = {
}

function meri.all_flags(f1, f2)
  f1 = f1 or ""
  f2 = f2 or ""
  for i = 1, #f1 do
    local f = f1:sub(i, i)
    if not f2:find(f) then
      return false
    end
  end
  return true
end

function meri.any_flags(f1, f2)
  f1 = f1 or ""
  f2 = f2 or ""
  for i = 1, #f1 do
    local f = f1:sub(i, i)
    if f2:find(f) then
      return true
    end
  end
  return false
end

function meri.lfrp_match(flags, tab)
  tab = tab or meri.self_metadata.lfrp_data or {}
  flags = flags or meri.self.flags
  return meri.all_flags(tab.required, flags) and not meri.any_flags(tab.excluded, flags)
end

function meri.channel_message_handler(event, from, mtype, channel, identifier, data)
  if identifier ~= 'libmeri_ping' and identifier ~= 'libmeri_ack' then
    -- not for us
    return
  end
  -- ignore own messages
  if slower(from) == meri.whoami then
    if not meri.lfrp_allow_self then
      return
    end
  end
  meri.debug("%s message from %s in %s.", identifier, from, channel)
  if identifier == 'libmeri_ack' then
    local bang = data:find('!')
    if not bang then
      meri.debug("Invalid string with no name! in libmeri_ack.")
      return
    end
    local name = slower(data:sub(1, bang - 1))
    if name ~= meri.whoami then
      -- for someone else
      return
    end
    meri.debug("ack from %s.", from)
    -- the part after the ! is supposed to be a table
    local message_tab = rar.from_data(data:sub(bang + 1, -1))
    if not message_tab or type(message_tab) ~= 'table' then
      meri.debug("Invalid libmeri_ack, expected table, got %s.",
        tostring(message_tab))
      return
    end
    if message_tab.data and slower(from) ~= meri.whoami then
      -- treat this as new incoming data
      meri.data[slower(from)] = { stamp = Inspect.Time.Server(), data = message_tab.data }
    end
    if not message_tab.handle then
      meri.debug("libmeri_ack with no handle, ignoring.")
      return
    end
    local existing = meri.lfrp_handles[message_tab.handle]
    if existing then
      existing.responses = existing.responses or {}
      -- store the channel info
      message_tab.channel = channel
      existing.responses[from] = message_tab
      meri.debug("Added response to handle %s for %s.", message_tab.handle, from)
      -- and we're done, because the responses are collated at the
      -- end of the timeout.
    end
  elseif identifier == 'libmeri_ping' then
    local message_tab = rar.from_data(data)
    if not message_tab or type(message_tab) ~= 'table' then
      meri.debug("Got an invalid message_tab: %s", tostring(message_tab))
      return
    end
    meri.debug("ping from %s: handle %s.", from, message_tab.handle)
    local unique = slower(from) .. '!' .. message_tab.handle
    if meri.seen_lfrp_handles[unique] then
      meri.debug("ignoring ping (already seen).")
      return
    else
      meri.seen_lfrp_handles[unique] = true
    end
    if meri.lfrp_match(nil, message_tab) then
      -- we match! let's respond.
      local message_new = { data = meri.self, lfrp_meta = tcopy(meri.self_metadata.lfrp_data), handle = message_tab.handle }
      -- are we in a party?
      local peeps = Inspect.Unit.List()
      for k, v in pairs(peeps) do
        if k:sub(1, 5) == 'group' then
	  message_new.lfrp_meta.party = true
	  break
	end
      end
      -- add location (maybe) and alliance:
      local details = Inspect.Unit.Detail('player')
      if message_new.lfrp_meta.show_location then
        message_new.lfrp_meta.zone = details.zone
        message_new.lfrp_meta.location = details.locationName
      end
      message_new.lfrp_meta.alliance = details.alliance
      local message_text = from .. '!' .. rar.to_data(message_new)
      -- don't bother with callback, we don't care.
      meri.debug("sending ack to %s.", channel)
      Command.Message.Broadcast('channel', channel, 'libmeri_ack', message_text)
    end
  end
end

function meri.lfrp_start_listening()
  Command.Message.Accept('channel', 'libmeri_ping')
  Command.Message.Accept('channel', 'libmeri_ack')
  local existing = Inspect.Event.List(Event.Message.Receive)
  -- add a new one only if there isn't one already
  for i = 1, #existing do
    if existing[i].label == 'libmerisioux_lfrp' then
      return
    end
  end
  Command.Event.Attach(Event.Message.Receive, meri.channel_message_handler, 'libmerisioux_lfrp')
end

function meri.lfrp_stop_listening()
  Command.Message.Reject('channel', 'libmeri_ping')
  Command.Message.Reject('channel', 'libmeri_ack')
  Command.Event.Detach(Event.Message.Receive, meri.channel_message_handler, 'libmerisioux_lfrp')
  -- clear out all handles
  meri.lfrp_handles = {}
end

function meri.lfrp_status(enable_disable)
  if enable_disable ~= nil then
    was_active = not not meri.self_metadata.lfrp_config.active
    meri.self_metadata.lfrp_config.active = not not enable_disable
    if meri.self_metadata.lfrp_config.active then
      meri.lfrp_start_listening()
    else
      meri.lfrp_stop_listening()
    end
  end
  return not not meri.self_metadata.lfrp_config.active
end

function meri.lfrp_channel(add_remove, channel)
  meri.self_metadata.lfrp_config.channels = meri.self_metadata.lfrp_config.channels or {}
  if channel then
    if add_remove == true then
      meri.self_metadata.lfrp_config.channels[channel] = true
    elseif add_remove == false then
      meri.self_metadata.lfrp_config.channels[channel] = nil
    end
  end
  return tkeys(meri.self_metadata.lfrp_config.channels)
end

function meri.lfrp_get(setting)
  if setting == 'comment' then
    return meri.self_metadata.lfrp_data.comment
  elseif setting == 'excluded' then
    return meri.self_metadata.lfrp_data.excluded
  elseif setting == 'required' then
    return meri.self_metadata.lfrp_data.required
  elseif setting == 'show_location' then
    return meri.self_metadata.lfrp_data.show_location
  end
  return nil
end

function meri.lfrp_set(setting, value)
  if setting == 'comment' then
    if value ~= nil then
      meri.self_metadata.lfrp_data.comment = tostring(value):sub(1, 80)
    else
      meri.self_metadata.lfrp_data.comment = nil
    end
  elseif setting == 'excluded' then
    meri.self_metadata.lfrp_data.excluded = meri.flags_from(tostring(value))
  elseif setting == 'required' then
    meri.self_metadata.lfrp_data.required = meri.flags_from(tostring(value))
  elseif setting == 'show_location' then
    meri.self_metadata.lfrp_data.show_location = not not value
  end
end

function meri.broadcast_callback(handle, failure, message)
  if not failure then
    return
  end
  if meri.lfrp_handles[handle] and meri.lfrp_handles[handle].callback then
    -- don't delete the handle, because a user might have had more than one
    -- channel requested.
    meri.lfrp_handles[handle].callback(handle, false, "Message broadcast failed: " .. (message or "no specific reason given."))
  end
end

function meri.lfrp_ping(callback, handle, timeout, ...)
  local channels = { ... }
  local now = Inspect.Time.Frame()
  if meri.lfrp_ping_stamp and (now - meri.lfrp_ping_stamp < 10) then
    callback(handle, false, "Error: Can't ping again this soon.")
    callback(handle, true, {})
    return
  else
    meri.lfrp_ping_stamp = now
  end
  if #channels < 1 then
    for k, v in pairs(meri.self_metadata.lfrp_config.channels) do
      channels[#channels + 1] = k
    end
    if #channels < 1 then
      callback(handle, false, "Error: No channels configured or requested.")
      callback(handle, true, {})
      return
    end
  end
  if not meri.self_metadata.lfrp_config.active then
    callback(handle, false, "Error: Requested a LibMerisioux LFRP ping, but LFRP is disabled.")
    callback(handle, true, {})
    return
  end
  if not timeout or type(timeout) ~= 'number' or timeout < 1 or timeout > 10 then
    timeout = 10
  end
  meri.lfrp_handles[handle] = { expire = Inspect.Time.Frame() + timeout, callback = callback, responses = {} }
  local message_tab = { handle = handle, required = meri.self_metadata.lfrp_data.required, excluded = meri.self_metadata.lfrp_data.excluded }
  local message_body = rar.to_data(message_tab)
  for i = 1, #channels do
    meri.debug("Sending ping to channel %s.", channels[i])
    Command.Message.Broadcast('channel', channels[i], 'libmeri_ping', message_body, function(...) meri.broadcast_callback(handle, ...) end)
  end
end

function meri.startup_tasks()
  if meri.self_metadata.lfrp_config.active then
    meri.lfrp_start_listening()
  end
  meri.update_hook_handle = Library.LibCron.new('LibMerisioux', 1, true, false, meri.update_hook)
end

Command.Event.Attach(Event.Addon.SavedVariables.Load.End, meri.variables_loaded, "variable loaded hook")
Command.Event.Attach(Event.Addon.Startup.End, meri.startup_tasks, "request timeouts")
Command.Event.Attach(Event.Addon.Shutdown.Begin, meri.clean_variables, "variable cleanup")
