--[[ Merisioux
     An addon for those special moments.

]]--

local addoninfo, meri = ...
local lmeri = Library.LibMerisioux
local stt = Library.LibScrollyTextThing
local strlower = string.lower
local strupper = string.upper
meri.version = "0.30-151128-15:56:24"
meri.tooltip_is_visible = false
local printf = Library.printf.printf
local printfh = Library.printf.printfh
local printfs = Library.printf.printfs
local printfhs = Library.printf.printfhs
local sprintf = Library.printf.sprintf
local max = math.max

local color_scale = 0.1
meri.clearer = { colorR = color_scale, colorB = color_scale, colorG = color_scale, strength = 2.5, blurX = 6, blurY = 6 }

local function capitalize(s)
  if not s or type(s) ~= 'string' then
    return 'Oops'
  end
  return strupper(s:sub(1, 1)) .. strlower(s:sub(2, -1))
end

-- capitalize multiple words
local function fancy_capitalize(s)
  if not s or type(s) ~= 'string' or #s == 0 then
    return 'Oops'
  end
  local remaining = s
  local done = nil
  while #remaining > 0 do
    local space = remaining:find(' ')
    if space then
      if not done then
        done = capitalize(remaining:sub(1, space - 1))
      else
        done = done .. ' ' .. capitalize(remaining:sub(1, space - 1))
      end
      remaining = remaining:sub(space + 1, -1)
      space = remaining:find('[^ ]')
      if not space then
        break
      end
      -- skip ahead to the first non-blank.
      if space > 1 then
        remaining = remaining:sub(space, -1)
      end
    else
      if not done then
        done = capitalize(remaining)
      else
        done = done .. ' ' .. capitalize(remaining)
      end
      break
    end
  end
  return done
end

function meri.help_message(frame, fmt, ...)
  if frame and frame.helptext then
    frame.helptext:SetText(sprintf(fmt, ...))
    if frame.helpsecondary then
      frame.helpsecondary:SetText("")
    end
  else
    printf(sprintf(fmt, ...))
  end
end

function meri.help_message2(frame, fmt, ...)
  if frame and frame.helpsecondary then
    frame.helpsecondary:SetText(sprintf(fmt, ...))
  else
    printf(sprintf(fmt, ...))
  end
end

function meri.dump(char, data, error)
  if data then
    printf("Got data for %s:", char or 'nil')
    dump(data)
  else
    printf("Error trying to get data for %s: %s", char or 'nil', error or 'nil')
  end
end

function meri.new_version_callback(version)
  -- ignore "dev" versions.
  if version:match("dev") then
    return
  end
  local comp = Library.LibVersion.compare(meri.sv.last_version_seen, version)
  if comp == -1 then
    printf("Saw newer version %s, you have %s.",
	version, meri.version)
    meri.sv.last_version_seen = version
  end
end

function meri.variables_loaded(event, addon)
  if addon == 'Merisioux' then
    meri.whoami = strlower(Inspect.Unit.Detail('player').name)
    MerisiouxUIData = MerisiouxUIData or { window_x = 150, window_y = 150 }
    meri.sv = MerisiouxUIData
    meri.sv.player_notes = meri.sv.player_notes or {}
    meri.sv.button_point = meri.sv.button_point or "CENTER"
    meri.sv.button_x = meri.sv.button_x or 70
    meri.sv.button_y = meri.sv.button_y or 60
    -- if the player moved this, let's put it somewhere else
    meri.place_button('player.target')
    meri.fields = lmeri.fields()
    meri.flags, meri.sorted_flags = lmeri.flags()
    if not meri.sv.last_version_seen then
      meri.sv.last_version_seen = meri.version
    else
      local comp = Library.LibVersion.compare(meri.sv.last_version_seen, meri.version)

      if comp == -1 then
        meri.sv.last_version_seen = meri.version
      elseif comp == 1 then
        printf("Merisioux:  Saw newer version %s, you have %s.",
	    meri.sv.last_version_seen, meri.version)
      end
    end
    lmeri.addon('Merisioux', meri.new_version_callback)
    printfh('<font color="#9900cc">The most perfect and ideal RP addon is now loaded!</font>')
    printfh('<font color="#9900cc">(Use /meri -e to edit, or /meri -h for help.)</font>')
    lmeri.snoop(Inspect.Unit.Detail('player').name, meri.load_my_data_callback)
  end
end

function meri.button_callback(unit, char, data, error)
  -- don't draw the button if we're no longer targeting that unit
  if data and char == meri.targeting and not (meri.sv and meri.sv.suppress_button) then
    meri.unitbuttons.buttons[unit]:SetVisible(true)
  else
    meri.unitbuttons.buttons[unit]:SetVisible(false)
  end
end

function meri.edit_callback(char, data, error)
  if not data or not next(data) then
    data = { error = error or "Unknown error." }
  end
  meri.run_editor(char, data)
end

function meri.load_my_data_callback(char, data, error)
  if data and next(data) then
    meri.my_name = char
    meri.my_local_data = data
    if not meri.editor then
      meri.create_editor()
    end
    if meri.editor and meri.editor.window:GetVisible() and meri.editor.savebutton:GetEnabled() then
      return
    end
    meri.populate_editor(char, data)
  end
end

--[[
  Okay, this gets complicated.
  ]]--
function meri.run_editor(char, data)
  if not meri.editor then
    meri.create_editor()
  end
  if not meri.editor then
    printf("Sorry, but I just... *can't*.")
    return
  end
  if meri.editor and meri.editor.window:GetVisible() and meri.editor.savebutton:GetEnabled() then
    meri.help_message(meri.editor, "Got a request to load new data, but you have unsaved changes.")
    return
  end
  meri.populate_editor(char, data)
  meri.show_editor()
end

-- generic handler
function meri.any_changed()
  meri.editor.savebutton:SetEnabled(true)
end

function meri.populate_editor(char, data)
  local is_me = char == meri.whoami

  meri.editor.edit_char = char

  meri.editor.prefix_field:SetVisible(is_me)
  meri.editor.suffix_field:SetVisible(is_me)
  meri.editor.override_field:SetVisible(is_me)
  meri.editor.title_field:SetVisible(is_me)

  meri.editor.age_field:SetVisible(is_me)
  meri.editor.height_field:SetVisible(is_me)
  meri.editor.weight_field:SetVisible(is_me)
  meri.editor.cgender_field:SetVisible(is_me)
  meri.editor.pgender_field:SetVisible(is_me)
  meri.editor.currently_field:SetVisible(is_me)

  meri.editor.age_display:SetVisible(not is_me)
  meri.editor.height_display:SetVisible(not is_me)
  meri.editor.weight_display:SetVisible(not is_me)
  meri.editor.cgender_display:SetVisible(not is_me)
  meri.editor.pgender_display:SetVisible(not is_me)
  meri.editor.currently_display:SetVisible(not is_me)

  -- the "Custom" label itself should go away.
  meri.editor.custom_label:SetVisible(is_me)
  meri.editor.custom_name_field:SetVisible(is_me)
  meri.editor.custom_data_field:SetVisible(is_me)
  meri.editor.custom_name_display:SetVisible(not is_me)
  meri.editor.custom_data_display:SetVisible(not is_me)

  meri.editor.prefix_label:SetVisible(is_me)
  meri.editor.suffix_label:SetVisible(is_me)
  meri.editor.override_label:SetVisible(is_me)
  meri.editor.title_label:SetVisible(is_me)

  meri.editor.note_label:SetVisible(not is_me)
  meri.editor.note_display:SetVisible(not is_me)

  meri.editor.description_place:SetVisible(is_me)
  meri.editor.biography_place:SetVisible(is_me)
  meri.editor.meta_place:SetVisible(is_me)

  local ybase = 230
  -- use full display area when not editing:
  if is_me then
    meri.editor.description_display_place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 5, ybase + 161)
    meri.editor.biography_display_place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 282, ybase + 161)
    meri.editor.meta_display_place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 559, ybase + 161)
  else
    meri.editor.description_display_place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 5, ybase + 20)
    meri.editor.biography_display_place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 282, ybase + 20)
    meri.editor.meta_display_place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 559, ybase + 20)
  end
  -- no editing_flags for now
  for k, v in pairs(meri.editor.flags_boxes) do
    if type(v) == 'table' and v.box then
      v.box:SetEnabled(is_me)
    end
  end

  meri.editor.prefix_field:SetText(data.prefix or '')
  meri.editor.override_field:SetText(data.override or '')
  meri.editor.suffix_field:SetText(data.suffix or '')
  meri.editor.title_field:SetText(data.title or '')

  meri.editor.age_field:SetText(data.age or '')
  meri.editor.height_field:SetText(data.height or '')
  meri.editor.weight_field:SetText(data.weight or '')
  meri.editor.cgender_field:SetText(data.cgender or '')
  meri.editor.pgender_field:SetText(data.pgender or '')
  meri.editor.currently_field:SetText(data.currently or '')
  meri.editor.custom_name_field:SetText(data.custom_name or '')
  meri.editor.custom_data_field:SetText(data.custom_data or '')
  meri.editor.age_display:SetText(data.age or '')
  meri.editor.note_display:SetText(meri.sv.player_notes[char] or '')
  meri.editor.height_display:SetText(data.height or '')
  meri.editor.weight_display:SetText(data.weight or '')
  meri.editor.cgender_display:SetText(data.cgender or '')
  meri.editor.pgender_display:SetText(data.pgender or '')
  meri.editor.currently_display:SetText(data.currently or '')
  if data.custom_name and #data.custom_name > 0 then
    meri.editor.custom_name_display:SetText(fancy_capitalize(data.custom_name) .. ':')
  else
    if data.custom_data and #data.custom_data > 0 then
      meri.editor.custom_name_display:SetText('Custom Field:')
    else
      meri.editor.custom_name_display:SetText('No custom field.')
    end
  end
  meri.editor.custom_data_display:SetText(data.custom_data or '')

  meri.editor.description_field:text(data.description or '')
  meri.editor.biography_field:text(data.biography or '')
  meri.editor.meta_field:text(data.meta or '')

  meri.editor.description_display:text(data.description or '')
  meri.editor.biography_display:text(data.biography or '')
  meri.editor.meta_display:text(data.meta or '')

  -- no editing flags yet
  meri.set_flag_boxes(meri.editor.flags_boxes, data.flags, is_me)

  meri.name_changed(nil, char)
  meri.title_changed()

  meri.help_message(meri.editor, '%s', data.message or (data.error and '(No RP data)') or '')

  if is_me then
    meri.editor.savebutton:SetEnabled(false)
    meri.editor.savebutton:SetVisible(true)
  else
    meri.editor.savebutton:SetEnabled(false)
    meri.editor.savebutton:SetVisible(false)
  end
end

function meri.set_flag_boxes(boxes, flags, is_me)
  boxes.editing_flags = nil
  boxes.flags = flags or ''
  for c, field in pairs(boxes) do
    if type(field) == 'table' and field.box then
      if flags and string.find(flags, c) then
        boxes[c].box:SetChecked(true)
        boxes[c].label:SetAlpha(1)
      else
        boxes[c].box:SetChecked(false)
        boxes[c].label:SetAlpha(is_me and 1 or 0.6)
      end
    end
  end
end

function meri.save_failed_callback(error)
  meri.help_message(meri.editor, "Save failed (probably data storage exceeded).")
  meri.help_message2(meri.editor, "Error was: %s", error)
end

function meri.save_changes()
  local updates = {}

  meri.editor.savebutton:SetEnabled(false)

  if meri.editor.edit_char ~= meri.whoami then
    meri.help_message(meri.editor, "Error:  Tried to save changes for '%s', but logged in as '%s'.",
    	meri.editor.edit_char, meri.whoami)
    return
  end
  updates.prefix = meri.editor.prefix_field:GetText()
  updates.override = meri.editor.override_field:GetText()
  updates.suffix = meri.editor.suffix_field:GetText()
  updates.title = meri.editor.title_field:GetText()

  updates.age = meri.editor.age_field:GetText()
  updates.height = meri.editor.height_field:GetText()
  updates.weight = meri.editor.weight_field:GetText()
  updates.cgender = meri.editor.cgender_field:GetText()
  updates.pgender = meri.editor.pgender_field:GetText()
  updates.currently = meri.editor.currently_field:GetText()
  updates.custom_name = meri.editor.custom_name_field:GetText()
  updates.custom_data = meri.editor.custom_data_field:GetText()

  updates.description = meri.editor.description_field:text()
  updates.biography = meri.editor.biography_field:text()
  updates.meta = meri.editor.meta_field:text()
  -- if editing_flags got set, save them, otherwise use existing
  updates.flags = meri.editor.flags_boxes.editing_flags or lmeri.self.flags
  local bytes_in, bytes_out, error = lmeri.flaunt(updates, nil, meri.save_failed_callback)

  if error then
    meri.help_message(meri.editor, "Save failed: %s", error)
    if bytes_in and bytes_out then
      meri.help_message2(meri.editor, "(%d bytes of storage, compressed to ~%d.)", bytes_in, bytes_out)
    end
  else
    if bytes_in then
      meri.help_message(meri.editor, "Save OK: %d bytes of storage, compressed to ~%d.", bytes_in, bytes_out)
    else
      meri.help_message(meri.editor, "%s", tostring(bytes_out))
      -- stash this data for later use
      lmeri.snoop(Inspect.Unit.Detail('player').name, meri.load_my_data_callback)
    end
  end
end

function meri.mousemove(window, x, y)
  if meri.sv then
    if window == meri.editor.window then
      meri.sv.window_x = x
      meri.sv.window_y = y
    elseif window == meri.lfrp.window then
      meri.sv.lfrp_x = x
      meri.sv.lfrp_y = y
    end
  end
end

function meri.description_changed(field, error)
  meri.editor.description_display:text(meri.editor.description_field:text())
  meri.any_changed()
  if error then
    meri.help_message(meri.editor, "%s", error)
  end
end

function meri.biography_changed(field, error)
  meri.editor.biography_display:text(meri.editor.biography_field:text())
  meri.any_changed()
  if error then
    meri.help_message(meri.editor, "%s", error)
  end
end

function meri.meta_changed(field, error)
  meri.editor.meta_display:text(meri.editor.meta_field:text())
  meri.any_changed()
  if error then
    meri.help_message(meri.editor, "%s", error)
  end
end

function meri.title_changed()
  if not meri.editor then
    return
  end
  local title = meri.editor.title_field:GetText() or ''

  if #title > meri.fields.title.max then
    title = string.sub(title, 1, meri.fields.title.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.title.max)
    meri.editor.title_field:SetText(title)
  end

  if title ~= '' then
    meri.editor.title_display:SetText('"' .. title .. '"')
  else
    meri.editor.title_display:SetText('')
  end
  meri.any_changed()
end

function meri.name_changed(event, char)
  if not meri.editor then
    return
  end
  local prefix = meri.editor.prefix_field:GetText()
  local override = meri.editor.override_field:GetText()
  local suffix = meri.editor.suffix_field:GetText()
  if #prefix > meri.fields.prefix.max then
    prefix = string.sub(prefix, 1, meri.fields.prefix.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.prefix.max)
    meri.editor.prefix_field:SetText(prefix)
  end

  if #override > meri.fields.override.max then
    override = string.sub(override, 1, meri.fields.override.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.override.max)
    meri.editor.override_field:SetText(override)
  end

  if #suffix > meri.fields.suffix.max then
    suffix = string.sub(suffix, 1, meri.fields.suffix.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.suffix.max)
    meri.editor.suffix_field:SetText(suffix)
  end

  local pretty = lmeri.prettyname(char or meri.whoami, { prefix = prefix, override = override, suffix = suffix })
  meri.editor.name_display:SetText(pretty)
  meri.any_changed()
end

function meri.stats_changed()
  if not meri.editor then
    return
  end
  local age = meri.editor.age_field:GetText()
  local height = meri.editor.height_field:GetText()
  local weight = meri.editor.weight_field:GetText()
  local cgender = meri.editor.cgender_field:GetText()
  local pgender = meri.editor.pgender_field:GetText()
  local currently = meri.editor.currently_field:GetText()
  local custom_name = meri.editor.custom_name_field:GetText()
  local custom_data = meri.editor.custom_data_field:GetText()

  if #age > meri.fields.age.max then
    age = string.sub(age, 1, meri.fields.age.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.age.max)
    meri.editor.age_field:SetText(age)
  end

  if #custom_name > meri.fields.custom_name.max then
    custom_name = string.sub(custom_name, 1, meri.fields.custom_name.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.custom_name.max)
    meri.editor.custom_name_field:SetText(custom_name)
  end

  if #custom_data > meri.fields.custom_data.max then
    custom_data = string.sub(custom_data, 1, meri.fields.custom_data.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.custom_data.max)
    meri.editor.custom_data_field:SetText(custom_data)
  end

  if #cgender > meri.fields.cgender.max then
    cgender = string.sub(cgender, 1, meri.fields.cgender.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.cgender.max)
    meri.editor.cgender_field:SetText(cgender)
  end

  if #pgender > meri.fields.pgender.max then
    pgender = string.sub(pgender, 1, meri.fields.pgender.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.pgender.max)
    meri.editor.pgender_field:SetText(pgender)
  end

  if #height > meri.fields.height.max then
    height = string.sub(height, 1, meri.fields.height.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.height.max)
    meri.editor.height_field:SetText(height)
  end

  if #weight > meri.fields.weight.max then
    weight = string.sub(weight, 1, meri.fields.weight.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.weight.max)
    meri.editor.weight_field:SetText(weight)
  end

  if #currently > meri.fields.currently.max then
    currently = string.sub(currently, 1, meri.fields.currently.max)
    meri.help_message(meri.editor, "There is a limit of %d characters.", meri.fields.currently.max)
    meri.editor.currently_field:SetText(currently)
  end
  meri.any_changed()
end

function meri.make_text(parent, x, y, text, point, frametype)
  local frame = UI.CreateFrame(frametype, 'Generic Label', parent)
  frame:SetPoint(point or "TOPLEFT", parent, "TOPLEFT", x, y)
  frame:SetText(text or '')
  return frame
end

function meri.make_field(parent, x, y, text, point)
  frame = meri.make_text(parent, x, y, text, point, 'RiftTextfield')
  frame:SetBackgroundColor(.1, .1, .1, 1)
  return frame
end

function meri.make_label(parent, x, y, text, point)
  local frame = meri.make_text(parent, x, y, text, point, 'Text')
  frame:SetFontSize(frame:GetFontSize() + 2)
  frame:SetFontColor(.75, .75, .55)
  frame:SetEffectGlow(meri.clearer)
  return frame
end

function meri.make_flag_checkbox(parent, x, y, c, helptext_container, handler)
  if not meri.flags or not meri.flags[c] then
    printf("Got request for flag checkbox for '%s', but no such flag.",
    	tostring(c))
    return
  end
  local frame = UI.CreateFrame('Frame', 'Checkbox ' .. c, parent)
  frame:SetPoint("TOPLEFT", parent, "TOPLEFT", x, y)
  frame:SetWidth(85)
  frame:SetHeight(22)
  local help_text = meri.flags[c].description
  local exclude_text = ''
  local excludes = {}
  for c2 in string.gmatch(meri.flags[c].excludes, '.') do
    if c2 and meri.flags[c2] then
      table.insert(excludes, meri.flags[c2].name)
    end
  end
  if #excludes > 0 then
    exclude_text = 'Excludes: ' .. table.concat(excludes, ', ')
  end
  local mousein = function()
    meri.helptext_stamp = Inspect.Time.Frame()
    helptext_container.helptext:SetText(help_text)
    helptext_container.helpsecondary:SetText(exclude_text)
  end
  local mouseout = function()
    if Inspect.Time.Frame() ~= meri.helptext_stamp then
      helptext_container.helptext:SetText('')
      helptext_container.helpsecondary:SetText('')
    end
  end
  frame.label = UI.CreateFrame('Text', 'Label ' .. c, frame)
  frame.box = UI.CreateFrame('RiftCheckbox', 'Label ' .. c, frame)
  frame.box:SetPoint("TOPRIGHT", frame, "TOPRIGHT")
  frame.label:SetPoint("TOPRIGHT", frame.box, "TOPLEFT", 0, -2)
  frame.label:SetText(meri.flags[c].name .. ':')
  frame.label:SetFontSize(frame.label:GetFontSize() + 1)
  frame.label:SetFontColor(.85, .85, .70)
  frame.label:SetEffectGlow(meri.clearer)
  frame.label:SetLayer(-2)
  frame.box:EventAttach(Event.UI.Checkbox.Change, function() handler(c) end, "check box")
  frame.box:SetLayer(0)
  frame:EventAttach(Event.UI.Input.Mouse.Cursor.In, mousein, "mouse in")
  frame:EventAttach(Event.UI.Input.Mouse.Cursor.Out, mouseout, "mouse out")
  frame.label:EventAttach(Event.UI.Input.Mouse.Cursor.In, mousein, "mouse in")
  frame.label:EventAttach(Event.UI.Input.Mouse.Cursor.Out, mouseout, "mouse out")
  frame.box:EventAttach(Event.UI.Input.Mouse.Cursor.In, mousein, "mouse in")
  frame.box:EventAttach(Event.UI.Input.Mouse.Cursor.Out, mouseout, "mouse out")
  return frame
end

function meri.check_box(frame, boxes, c, set, verbose)
  if not frame then
    return
  end
  if boxes and boxes[c] then
    if set == nil then
      set = boxes[c].box:GetChecked()
    else
      boxes[c].box:SetChecked(set and true or false)
    end
    local removed, new_flags = lmeri.setflag(c, set, boxes.editing_flags or boxes.flags)
    -- stash a list of the flags under consideration
    boxes.editing_flags = new_flags
    local existing = lmeri.flags()[c]
    if existing and verbose then
      if set then
        printf("Set flag: %s", existing.description)
      else
        printf("Cleared flag: %s", existing.description)
      end
    end
    if removed and #removed > 0 then
      for i = 1, #removed do
        local c2 = string.sub(removed, i, i)
	if boxes[c2] then
	  boxes[c2].box:SetChecked(false)
	else
	  printf("Was asked to remove unknown flag %s.", c2)
	end
      end
    end
    -- and now the frame-specific stuff
    if frame == meri.editor then
      meri.any_changed()
    end
    if frame == meri.lfrp then
      meri.lfrp_changed()
    end
  else
    printf("Checked: Unknown flag %s", c)
  end
end

function meri.make_help_label(parent, x, y, text, point, helptext)
  local content = parent.content
  local capitalized = capitalize(text)
  local frame = meri.make_label(content, x, y, capitalized .. ":", point)
  helptext = helptext or (meri.fields[text] and meri.fields[text].help)
  helptext = helptext or "No help available."
  local mousein = function()
    meri.helptext_stamp = Inspect.Time.Frame()
    parent.helptext:SetText(helptext)
    parent.helpsecondary:SetText('')
  end
  local mouseout = function()
    if Inspect.Time.Frame() ~= meri.helptext_stamp then
      parent.helptext:SetText('')
    end
  end
  frame:EventAttach(Event.UI.Input.Mouse.Cursor.In, mousein, "mouse in")
  frame:EventAttach(Event.UI.Input.Mouse.Cursor.Out, mouseout, "mouse out")
  return frame
end

function meri.tab_to_field(field, event, key)
  if key == 'Tab' and field.nextfield then
    field.nextfield:SetKeyFocus(true)
  end
end

function meri.create_lfrp()
  local active = lmeri.lfrp_status()

  meri.lfrp = {}
  meri.lfrp.ctx = UI.CreateContext("Merisioux LFRP")
  if not meri.lfrp.ctx then
    printf("I don't think I can make a UI.")
    meri.lfrp = nil
    return
  end
  meri.lfrp.window = UI.CreateFrame('RiftWindow', 'LFRP', meri.lfrp.ctx)
  meri.lfrp.window:SetTitle("Any RP happening?")
  if meri.sv and meri.sv.lfrp_x then
    meri.lfrp.window:SetPoint("TOPLEFT", UIParent, "TOPLEFT",
    	meri.sv.lfrp_x, meri.sv.lfrp_y)
  end
  meri.lfrp.window:SetHeight(670)
  meri.lfrp.window:SetWidth(610)

  local l, t, r, b = meri.lfrp.window:GetTrimDimensions()
  r = r * -1
  b = b * -1

  Library.LibDraggable.draggify(meri.lfrp.window, meri.mousemove)
  meri.lfrp.closebutton = UI.CreateFrame("RiftButton", "Merisioux", meri.lfrp.window)
  meri.lfrp.closebutton:SetSkin("close")
  meri.lfrp.closebutton:SetPoint("TOPRIGHT", meri.lfrp.window, "TOPRIGHT", r + 5, 17)
  meri.lfrp.closebutton:EventAttach(Event.UI.Input.Mouse.Left.Click, meri.hide_lfrp, "close")
  meri.lfrp.content = meri.lfrp.window:GetContent()

  meri.lfrp.status_display = meri.make_label(meri.lfrp.content, 295, -8, "", "TOPCENTER")
  meri.lfrp.status_display:SetFontColor(0.9, 0.9, 0.9)
  meri.lfrp.status_display:SetFontSize(meri.lfrp.status_display:GetFontSize() + 2)
  meri.lfrp.status_display:SetText(active and "LFRP: Active" or "LFRP: Inactive")
  meri.lfrp.status_display:SetEffectGlow(meri.clearer)

  meri.lfrp.helptext = meri.make_label(meri.lfrp.content, 5, 200)
  meri.lfrp.helptext:SetFontColor(0.8, 0.8, 0.8)
  meri.lfrp.helptext:SetWidth(480)
  meri.lfrp.helptext:SetEffectGlow(meri.clearer)

  meri.lfrp.helpsecondary = meri.make_label(meri.lfrp.content, 5, 218)
  meri.lfrp.helpsecondary:SetFontColor(0.8, 0.8, 0.8)
  meri.lfrp.helpsecondary:SetWidth(480)
  meri.lfrp.helpsecondary:SetEffectGlow(meri.clearer)

  local ybase = 20
  meri.lfrp.comment_label = meri.make_help_label(meri.lfrp, 74, ybase + 0, "comment", "TOPRIGHT", "Comment to display in search results.")
  meri.lfrp.comment_field = meri.make_field(meri.lfrp.content, 75, ybase + 1)
  meri.lfrp.comment_field:SetWidth(460)
  meri.lfrp.comment_field:EventAttach(Event.UI.Textfield.Change, function() meri.lfrp_changed() end, "lfrp comment change")
  meri.lfrp.comment_display = meri.make_label(meri.lfrp.content, 65, ybase + 0)
  meri.lfrp.comment_display:SetWidth(460)
  meri.lfrp.comment_display:SetFontColor(.9, .9, .9)
  meri.lfrp.comment_display:SetVisible(false)

  meri.make_flag_checkboxes(meri.lfrp, 43, 'required', "Flags other players must have.")
  meri.make_flag_checkboxes(meri.lfrp, 115, 'excluded', "Flags other players must not have.")

  ybase = 182
  meri.lfrp.channel_label = meri.make_help_label(meri.lfrp, 74, ybase, "channels", "TOPRIGHT", "Channels used for pings.")
  meri.lfrp.channel_display = meri.make_label(meri.lfrp.content, 75, ybase)

  meri.lfrp.savebutton = UI.CreateFrame('RiftButton', 'Save Button', meri.lfrp.content)
  meri.lfrp.savebutton:SetPoint('TOPRIGHT', meri.lfrp.content, 'TOPRIGHT', -70, -12)
  meri.lfrp.savebutton:SetText('SAVE')
  meri.lfrp.savebutton:SetEnabled(false)
  meri.lfrp.savebutton:EventAttach(Event.UI.Input.Mouse.Left.Click, meri.lfrp_save_changes, "save button")

  meri.lfrp.togglebutton = UI.CreateFrame('RiftButton', 'Toggle Button', meri.lfrp.content)
  meri.lfrp.togglebutton:SetPoint('TOPLEFT', meri.lfrp.content, 'TOPLEFT', 62, -12)
  meri.lfrp.togglebutton:SetEnabled(true)
  meri.lfrp.togglebutton:SetText(active and 'DISABLE' or 'ENABLE')
  meri.lfrp.togglebutton:EventAttach(Event.UI.Input.Mouse.Left.Click, meri.lfrp_toggle, "toggle button")

  meri.lfrp.pingbutton = UI.CreateFrame('RiftButton', 'Ping Button', meri.lfrp.content)
  meri.lfrp.pingbutton:SetPoint('TOPRIGHT', meri.lfrp.content, 'TOPRIGHT', -5, 205)
  meri.lfrp.pingbutton:SetText('PING')
  meri.lfrp.pingbutton:SetEnabled(lmeri.lfrp_status())
  meri.lfrp.pingbutton:EventAttach(Event.UI.Input.Mouse.Left.Click, meri.lfrp_ping, "ping button")

  meri.lfrp.results_label = meri.make_label(meri.lfrp.content, 5, 245, 'Results:', 'TOPLEFT')
  meri.lfrp.results_label:SetFontSize(meri.lfrp.results_label:GetFontSize() + 2)

  meri.lfrp.results_display = UI.CreateFrame('Text', 'lfrp_results', meri.lfrp.content)
  meri.lfrp.results_display:SetFontSize(meri.lfrp.results_display:GetFontSize() + 2)
  meri.lfrp.results_display:SetPoint('TOPLEFT', meri.lfrp.content, 'TOPLEFT', 10, 268)
  meri.lfrp.results_display:SetPoint('BOTTOMRIGHT', meri.lfrp.content, 'BOTTOMRIGHT', -10, -5)

  meri.lfrp.window:SetVisible(false)
end

function meri.lfrp_changed()
  local comment = meri.lfrp.comment_field:GetText() or ""
  if #comment > 80 then
    meri.lfrp.comment_field:SetText(comment:sub(1, 80))
    meri.help_message(meri.lfrp, "Comments are restricted to 80 characters.")
  end
  meri.lfrp.savebutton:SetEnabled(true)
end

meri.lfrp_ping_counter = 0

-- sorry about the namespace pollution
function meri_snoop_callback(name)
  lmeri.snoop(name, meri.edit_callback)
end

function meri.lfrp_ping_callback(handle, status, details)
  if tonumber(handle) == meri.lfrp_ping_counter then
    if status then
      local lines = {}
      for k, v in pairs(details) do
	-- special case for debugging: allow self.
	if strlower(k) ~= meri.whoami or lmeri.lfrp_allow_self then
	  local comment = v.lfrp_meta and v.lfrp_meta.comment or "(No comment.)"
	  local location = ""
	  if v.lfrp_meta.zone then
	    local details = Inspect.Zone.Detail(v.lfrp_meta.zone)
	    location = sprintf("(%s) ", details.name)
	  end
	  local pretty_print = sprintf("<font color=\"#aaaaff\"><a lua=\"meri_snoop_callback('%s')\">%s</a></font><font color=\"#cccccc\">: %s%s</font>", k, k, location, comment)
	  lines[#lines + 1] = pretty_print
        end
      end
      if #lines == 0 then
        lines[1] = "<font color=\"#cccc66\">No results.</font>"
	if meri.lfrp_error then
          lines[2] = "<font color=\"#cccc66\">" .. meri.lfrp_error .. "</font>"
	end
      end
      if #lines > 16 then
        lines[16] = sprintf("<font color=\"#aaaaaa\">[... and %d more.]</font>", #lines - 16)
      end
      local fancy_text = table.concat(lines, "\n")
      meri.lfrp.results_display:SetText(fancy_text, true)
    else
      meri.lfrp_error = details
      meri.help_message(meri.lfrp, "Ping: %s", details)
    end
  end
end

function meri.lfrp_ping()
  meri.lfrp_ping_counter = meri.lfrp_ping_counter + 1
  local timer = 2
  meri.help_message(meri.lfrp, "Sending ping (wait about %d seconds).", timer)
  meri.lfrp_error = nil
  meri.lfrp.results_display:SetText("<font color=\"#aaaaaa\">Waiting...</font>", true)
  lmeri.lfrp_ping(meri.lfrp_ping_callback, tostring(meri.lfrp_ping_counter), timer)
end

function meri.lfrp_toggle()
  local active = lmeri.lfrp_status()
  active = lmeri.lfrp_status(not active)
  meri.lfrp.togglebutton:SetText(active and 'DISABLE' or 'ENABLE')
  meri.lfrp.pingbutton:SetEnabled(active)
  meri.lfrp.status_display:SetText(active and "LFRP: Active" or "LFRP: Inactive")
  active = active and "active" or "inactive"
  printf("LFRP now %s.", active)
end

function meri.lfrp_save_changes()
  if meri.lfrp.required_boxes.editing_flags then
    lmeri.lfrp_set('required', meri.lfrp.required_boxes.editing_flags)
  end
  if meri.lfrp.excluded_boxes.editing_flags then
    lmeri.lfrp_set('excluded', meri.lfrp.excluded_boxes.editing_flags)
  end
  local comment = meri.lfrp.comment_field:GetText()
  if comment and #comment > 0 then
    lmeri.lfrp_set('comment', comment)
  else
    lmeri.lfrp_set('comment', nil)
  end
  local status = lmeri.lfrp_status() and "active" or "inactive"
  printf("Saved LFRP data. LFRP %s. Comment: %s.", status, (#comment > 0) and comment or "<none>")
  meri.lfrp.savebutton:SetEnabled(false)
end

function meri.run_lfrp()
  if not meri.lfrp then
    meri.create_lfrp()
    if not meri.lfrp then
      return
    end
  end
  local excluded = lmeri.lfrp_get('excluded') or ""
  local required = lmeri.lfrp_get('required') or ""
  local comment = lmeri.lfrp_get('comment') or ""
  local show_location = lmeri.lfrp_get('show_location') or false
  meri.set_flag_boxes(meri.lfrp.excluded_boxes, excluded, true)
  meri.set_flag_boxes(meri.lfrp.required_boxes, required, true)
  meri.lfrp.comment_field:SetText(comment)
  meri.lfrp.savebutton:SetEnabled(false)
  meri.lfrp_update_channel_display()
  meri.show_lfrp()
end

function meri.lfrp_update_channel_display()
  if meri.lfrp and meri.lfrp.channel_display then
    local channels
    local listing = table.concat(lmeri.lfrp_channel())
    if #listing > 0 then
      meri.lfrp.channel_display:SetText(listing)
    else
      meri.lfrp.channel_display:SetText("No channels selected! /meri -R <channel> to add.")
    end
  end
end

function meri.show_lfrp()
  if meri.lfrp and meri.lfrp.window then
    meri.lfrp.window:SetVisible(true)
  end
end

function meri.hide_lfrp()
  if meri.lfrp and meri.lfrp.window then
    meri.lfrp.window:SetVisible(false)
    meri.lfrp.comment_field:SetKeyFocus(false)
  end
end

function meri.create_editor()
  meri.editor = {}
  meri.editor.ctx = UI.CreateContext("Merisioux Editor")
  if not meri.editor.ctx then
    printf("I don't think I can make a UI.")
    meri.editor = nil
    return
  end
  meri.editor.window = UI.CreateFrame('RiftWindow', 'Editor', meri.editor.ctx)
  meri.editor.window:SetTitle("Let Me Tell You About My Backstory...")
  if meri.sv and meri.sv.window_x then
    meri.editor.window:SetPoint("TOPLEFT", UIParent, "TOPLEFT",
    	meri.sv.window_x, meri.sv.window_y)
  end
  meri.editor.window:SetHeight(670)
  meri.editor.window:SetWidth(860)
  local l, t, r, b = meri.editor.window:GetTrimDimensions()
  r = r * -1
  b = b * -1
  Library.LibDraggable.draggify(meri.editor.window, meri.mousemove)

  meri.editor.closebutton = UI.CreateFrame("RiftButton", "Merisioux", meri.editor.window)
  meri.editor.closebutton:SetSkin("close")
  meri.editor.closebutton:SetPoint("TOPRIGHT", meri.editor.window, "TOPRIGHT", r + 5, 17)
  meri.editor.closebutton:EventAttach(Event.UI.Input.Mouse.Left.Click, meri.hide_editor, "close")

  meri.editor.content = meri.editor.window:GetContent()

  meri.editor.helptext = meri.make_label(meri.editor.content, 5, 562)
  meri.editor.helptext:SetFontColor(0.8, 0.8, 0.8)
  meri.editor.helptext:SetWidth(730)
  meri.editor.helptext:SetEffectGlow(meri.clearer)

  meri.editor.helpsecondary = meri.make_label(meri.editor.content, 5, 578)
  meri.editor.helpsecondary:SetFontColor(0.8, 0.8, 0.8)
  meri.editor.helpsecondary:SetWidth(730)
  meri.editor.helpsecondary:SetEffectGlow(meri.clearer)

  local lab, place

  meri.editor.name_display = meri.make_label(meri.editor.content, 419, -8, "", "TOPCENTER")
  meri.editor.name_display:SetFontSize(meri.editor.name_display:GetFontSize() + 2)
  meri.editor.name_display:SetText("Just Testing")
  meri.editor.name_display:SetEffectGlow(meri.clearer)

  meri.editor.title_display = meri.make_label(meri.editor.content, 419, 12, "", "TOPCENTER")
  meri.editor.title_display:SetFontSize(meri.editor.title_display:GetFontSize() + 1)
  meri.editor.title_display:SetText("Just Testing")
  meri.editor.title_display:SetFontColor(0.85, 0.85, 0.45)
  meri.editor.title_display:SetEffectGlow(meri.clearer)

  local ybase = 34

  meri.editor.note_label = meri.make_help_label(meri.editor, 67, ybase + 0, "note", "TOPRIGHT", "Use /meri -n <name> <note> to update.")
  meri.editor.note_label:SetVisible(false)
  meri.editor.note_display = meri.make_label(meri.editor.content, 68, ybase + 0)
  meri.editor.note_display:SetWidth(610)
  meri.editor.note_display:SetFontColor(.9, .9, .9)
  meri.editor.note_display:SetVisible(false)

  meri.editor.prefix_label = meri.make_help_label(meri.editor, 67, ybase + 0, "prefix", "TOPRIGHT")
  meri.editor.prefix_field = meri.make_field(meri.editor.content, 68, ybase + 1)
  meri.editor.prefix_field:SetWidth(210)
  meri.editor.prefix_field:EventAttach(Event.UI.Textfield.Change, function() meri.name_changed() end, "prefix change")

  meri.editor.override_label = meri.make_help_label(meri.editor, 342, ybase + 0, "override", "TOPRIGHT")
  meri.editor.override_field = meri.make_field(meri.editor.content, 343, ybase + 1)
  meri.editor.override_field:SetWidth(210)
  meri.editor.override_field:EventAttach(Event.UI.Textfield.Change, function() meri.name_changed() end, "override change")

  meri.editor.suffix_label = meri.make_help_label(meri.editor, 607, ybase + 0, "suffix", "TOPRIGHT")
  meri.editor.suffix_field = meri.make_field(meri.editor.content, 608, ybase + 1)
  meri.editor.suffix_field:SetWidth(210)
  meri.editor.suffix_field:EventAttach(Event.UI.Textfield.Change, function() meri.name_changed() end, "suffix change")

  meri.editor.title_label = meri.make_help_label(meri.editor, 214, ybase + 23, "title", "TOPRIGHT")
  meri.editor.title_field = meri.make_field(meri.editor.content, 215, ybase + 24)
  meri.editor.title_field:SetWidth(420)
  meri.editor.title_field:EventAttach(Event.UI.Textfield.Change, meri.title_changed, "title change")

  -- age/height/weight
  ybase = 80
  meri.editor.age_label = meri.make_help_label(meri.editor, 67, ybase + 0, "age", "TOPRIGHT")
  meri.editor.age_field = meri.make_field(meri.editor.content, 68, ybase + 1)
  meri.editor.age_field:SetWidth(210)
  meri.editor.age_field:EventAttach(Event.UI.Textfield.Change, function() meri.stats_changed() end, "age change")
  meri.editor.age_display = meri.make_label(meri.editor.content, 68, ybase + 0)
  meri.editor.age_display:SetWidth(210)
  meri.editor.age_display:SetFontColor(.9, .9, .9)
  meri.editor.age_display:SetVisible(false)

  meri.editor.height_label = meri.make_help_label(meri.editor, 342, ybase + 0, "height", "TOPRIGHT")
  meri.editor.height_field = meri.make_field(meri.editor.content, 343, ybase + 1)
  meri.editor.height_field:SetWidth(210)
  meri.editor.height_field:EventAttach(Event.UI.Textfield.Change, function() meri.stats_changed() end, "height change")
  meri.editor.height_display = meri.make_label(meri.editor.content, 343, ybase + 0)
  meri.editor.height_display:SetWidth(210)
  meri.editor.height_display:SetFontColor(.9, .9, .9)
  meri.editor.height_display:SetVisible(false)

  meri.editor.weight_label = meri.make_help_label(meri.editor, 607, ybase + 0, "weight", "TOPRIGHT")
  meri.editor.weight_field = meri.make_field(meri.editor.content, 608, ybase + 1)
  meri.editor.weight_field:SetWidth(210)
  meri.editor.weight_field:EventAttach(Event.UI.Textfield.Change, function() meri.stats_changed() end, "weight change")
  meri.editor.weight_display = meri.make_label(meri.editor.content, 608, ybase + 0)
  meri.editor.weight_display:SetWidth(210)
  meri.editor.weight_display:SetFontColor(.9, .9, .9)
  meri.editor.weight_display:SetVisible(false)

  -- genders
  ybase = 103
  meri.editor.cgender_label = meri.make_help_label(meri.editor, 156, ybase + 0, "gender", "TOPRIGHT")
  meri.editor.cgender_field = meri.make_field(meri.editor.content, 155, ybase + 1)
  meri.editor.cgender_field:SetWidth(210)
  meri.editor.cgender_field:EventAttach(Event.UI.Textfield.Change, function() meri.stats_changed() end, "cgender change")
  meri.editor.cgender_display = meri.make_label(meri.editor.content, 155, ybase + 0)
  meri.editor.cgender_display:SetWidth(210)
  meri.editor.cgender_display:SetFontColor(.9, .9, .9)
  meri.editor.cgender_display:SetVisible(false)

  meri.editor.pgender_label = meri.make_help_label(meri.editor, 499, ybase + 0, "gender (player)", "TOPRIGHT")
  meri.editor.pgender_field = meri.make_field(meri.editor.content, 500, ybase + 1)
  meri.editor.pgender_field:SetWidth(210)
  meri.editor.pgender_field:EventAttach(Event.UI.Textfield.Change, function() meri.stats_changed() end, "pgender change")
  meri.editor.pgender_display = meri.make_label(meri.editor.content, 500, ybase + 0)
  meri.editor.pgender_display:SetWidth(210)
  meri.editor.pgender_display:SetFontColor(.9, .9, .9)
  meri.editor.pgender_display:SetVisible(false)

  ybase = 126
  meri.editor.currently_label = meri.make_help_label(meri.editor, 67, ybase + 0, "currently", "TOPRIGHT")
  meri.editor.currently_field = meri.make_field(meri.editor.content, 68, ybase + 1)
  meri.editor.currently_field:SetWidth(610)
  meri.editor.currently_field:EventAttach(Event.UI.Textfield.Change, function() meri.stats_changed() end, "currently change")
  meri.editor.currently_display = meri.make_label(meri.editor.content, 68, ybase + 0)
  meri.editor.currently_display:SetWidth(610)
  meri.editor.currently_display:SetFontColor(.9, .9, .9)
  meri.editor.currently_display:SetVisible(false)

  ybase = 149
  meri.editor.custom_label = meri.make_help_label(meri.editor, 67, ybase + 0, "custom", "TOPRIGHT", "Custom field (name your own!)")
  meri.editor.custom_name_field = meri.make_field(meri.editor.content, 68, ybase + 1)
  meri.editor.custom_name_field:SetWidth(180)
  meri.editor.custom_name_field:EventAttach(Event.UI.Textfield.Change, function() meri.stats_changed() end, "custom change")
  meri.editor.custom_name_display = meri.make_label(meri.editor.content, 1, ybase - 1, "custom")
  meri.editor.custom_name_display:SetVisible(false)

  meri.editor.custom_data_field = meri.make_field(meri.editor.content, 255, ybase + 1)
  meri.editor.custom_data_field:SetWidth(510)
  meri.editor.custom_data_field:EventAttach(Event.UI.Textfield.Change, function() meri.stats_changed() end, "custom change")
  meri.editor.custom_data_display = meri.make_label(meri.editor.content, 255, ybase + 0)
  meri.editor.custom_data_display:SetPoint("LEFT", meri.editor.custom_name_display, "RIGHT", 1, nil)
  meri.editor.custom_data_display:SetWidth(510)
  meri.editor.custom_data_display:SetFontColor(.9, .9, .9)
  meri.editor.custom_data_display:SetVisible(false)

  ybase = 230
  lab = meri.make_help_label(meri.editor, 5, ybase + 0, "description", "TOPLEFT")
  place = UI.CreateFrame('Frame', 'Layout Frame', meri.editor.content)
  place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 5, ybase + 20)
  place:SetPoint("BOTTOMRIGHT", meri.editor.content, "TOPLEFT", 275, ybase + 150)
  meri.editor.description_place = place
  meri.editor.description_field = stt.create(place, "Merisioux", '', { wordwrap = true, editable = true, max = meri.fields.description.max }, 'RIGHT', meri.description_changed) 
  meri.editor.description_field.background:SetBackgroundColor(0.1, 0.1, 0.1, 0.9)

  place = UI.CreateFrame('Frame', 'Layout Frame', meri.editor.content)
  place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 5, ybase + 161)
  place:SetPoint("BOTTOMRIGHT", meri.editor.content, "TOPLEFT", 275, ybase + 330)
  meri.editor.description_display_place = place
  meri.editor.description_display = stt.create(place, "Merisioux", '', { wordwrap = true, size = 14, glow = meri.clearer }, 'RIGHT')

  lab = meri.make_help_label(meri.editor, 282, ybase + 0, "biography", "TOPLEFT")
  place = UI.CreateFrame('Frame', 'Layout Frame', meri.editor.content)
  place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 282, ybase + 20)
  place:SetPoint("BOTTOMRIGHT", meri.editor.content, "TOPLEFT", 552, ybase + 150)
  meri.editor.biography_place = place
  meri.editor.biography_field = stt.create(place, "Merisioux", '', { wordwrap = true, editable = true, max = meri.fields.biography.max }, 'RIGHT', meri.biography_changed) 
  meri.editor.biography_field.background:SetBackgroundColor(0.1, 0.1, 0.1, 0.9)

  place = UI.CreateFrame('Frame', 'Layout Frame', meri.editor.content)
  place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 282, ybase + 161)
  place:SetPoint("BOTTOMRIGHT", meri.editor.content, "TOPLEFT", 552, ybase + 330)
  meri.editor.biography_display_place = place
  meri.editor.biography_display = stt.create(place, "Merisioux", '', { wordwrap = true, size = 14, glow = meri.clearer }, 'RIGHT')

  lab = meri.make_help_label(meri.editor, 559, ybase + 0, "meta", "TOPLEFT")
  place = UI.CreateFrame('Frame', 'Layout Frame', meri.editor.content)
  place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 559, ybase + 20)
  place:SetPoint("BOTTOMRIGHT", meri.editor.content, "TOPLEFT", 829, ybase + 150)
  meri.editor.meta_place = place
  meri.editor.meta_field = stt.create(place, "Merisioux", '', { wordwrap = true, editable = true, max = meri.fields.meta.max }, 'RIGHT', meri.meta_changed) 
  meri.editor.meta_field.background:SetBackgroundColor(0.1, 0.1, 0.1, 0.9)

  place = UI.CreateFrame('Frame', 'Layout Frame', meri.editor.content)
  place:SetPoint("TOPLEFT", meri.editor.content, "TOPLEFT", 559, ybase + 161)
  place:SetPoint("BOTTOMRIGHT", meri.editor.content, "TOPLEFT", 829, ybase + 330)
  meri.editor.meta_display_place = place
  meri.editor.meta_display = stt.create(place, "Merisioux", '', { wordwrap = true, size = 14, glow = meri.clearer }, 'RIGHT')

  meri.editor.savebutton = UI.CreateFrame('RiftButton', 'Save Button', meri.editor.content)
  meri.editor.savebutton:SetPoint('BOTTOMRIGHT', meri.editor.content, 'BOTTOMRIGHT', -5, -5)
  meri.editor.savebutton:SetText('SAVE')
  meri.editor.savebutton:SetEnabled(false)
  meri.editor.savebutton:EventAttach(Event.UI.Input.Mouse.Left.Click, meri.save_changes, "save button")

  meri.editor.lfrpbutton = UI.CreateFrame('RiftButton', 'Lfrp Button', meri.editor.content)
  meri.editor.lfrpbutton:SetPoint('BOTTOMRIGHT', meri.editor.content, 'BOTTOMRIGHT', -135, -5)
  meri.editor.lfrpbutton:SetText('LFRP')
  meri.editor.lfrpbutton:SetEnabled(true)
  meri.editor.lfrpbutton:EventAttach(Event.UI.Input.Mouse.Left.Click, function() meri.run_lfrp() end, "lfrp button")

  meri.editor.prefix_field.nextfield = meri.editor.override_field
  meri.editor.override_field.nextfield = meri.editor.suffix_field
  meri.editor.suffix_field.nextfield = meri.editor.title_field
  meri.editor.title_field.nextfield = meri.editor.age_field
  meri.editor.age_field.nextfield = meri.editor.height_field
  meri.editor.height_field.nextfield = meri.editor.weight_field
  meri.editor.weight_field.nextfield = meri.editor.cgender_field
  meri.editor.cgender_field.nextfield = meri.editor.pgender_field
  meri.editor.pgender_field.nextfield = meri.editor.currently_field
  meri.editor.currently_field.nextfield = meri.editor.custom_name_field
  meri.editor.custom_name_field.nextfield = meri.editor.custom_data_field
  meri.editor.custom_data_field.nextfield = meri.editor.description_field
  meri.editor.description_field.nextfield = meri.editor.biography_field
  meri.editor.biography_field.nextfield = meri.editor.meta_field
  meri.editor.meta_field.nextfield = meri.editor.prefix_field

  meri.editor.prefix_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.override_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.suffix_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.title_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.age_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.height_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.weight_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.cgender_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.pgender_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.currently_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.custom_name_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")
  meri.editor.custom_data_field:EventAttach(Event.UI.Input.Key.Down, meri.tab_to_field, "tab")

  meri.make_flag_checkboxes(meri.editor, 173, 'flags', "Flags describing your character/RP.")
  meri.editor.window:SetVisible(false)
end

function meri.make_flag_checkboxes(frame, ybase, heading, helptext)
  lab = meri.make_help_label(frame, 64, ybase - 2, heading, "TOPRIGHT", helptext)
  local boxname = heading .. '_boxes'
  frame[boxname] = {}
  local boxes = frame[boxname]
  for i = 1, #meri.sorted_flags do
    local c = string.sub(meri.sorted_flags, i, i)
    local loc = i - 1
    -- horrid hack:  bump lore/non-lore to a new column
    if loc >= 11 then
      loc = loc + 1
    end
    -- and again for the conflict-resolution flags...
    if loc >= 13 then
      loc = loc + 1
    end
    boxes[c] = meri.make_flag_checkbox(frame.content, 65 + 85 * math.floor(loc / 3), ybase + 19 * (loc % 3), c, frame, function(...) return meri.check_box(frame, boxes, ...) end)
  end
end

function meri.show_editor()
  if meri.editor and meri.editor.window then
    meri.editor.window:SetVisible(true)
  end
end

function meri.hide_editor()
  if meri.editor and meri.editor.window then
    meri.editor.window:SetVisible(false)
    meri.editor.prefix_field:SetKeyFocus(false)
    meri.editor.override_field:SetKeyFocus(false)
    meri.editor.suffix_field:SetKeyFocus(false)
    meri.editor.title_field:SetKeyFocus(false)
    meri.editor.age_field:SetKeyFocus(false)
    meri.editor.height_field:SetKeyFocus(false)
    meri.editor.weight_field:SetKeyFocus(false)
    meri.editor.cgender_field:SetKeyFocus(false)
    meri.editor.pgender_field:SetKeyFocus(false)
    meri.editor.currently_field:SetKeyFocus(false)
    meri.editor.custom_name_field:SetKeyFocus(false)
    meri.editor.custom_data_field:SetKeyFocus(false)
    meri.editor.description_field:SetKeyFocus(false)
    meri.editor.biography_field:SetKeyFocus(false)
    meri.editor.meta_field:SetKeyFocus(false)
  end
end

local valid_point_names = {
  TOPLEFT = true,
  TOPCENTER = true,
  TOPRIGHT = true,
  CENTERLEFT = true,
  CENTER = true,
  CENTERRIGHT = true,
  BOTTOMLEFT = true,
  BOTTOMCENTER = true,
  BOTTOMRIGHT = true,
}

function meri.slashcommand(args)
  local did_something = false
  local replace_button = false
  if not args then
    return
  end
  if args.r then
    meri.run_lfrp()
    did_something = true
  end
  if args.n then
    local name = strlower(args.n)
    local capname = capitalize(args.n)
    did_something = true
    if #args.leftover_full > 0 then
      if args.leftover_full == '~' then
        meri.sv.player_notes[name] = nil
        printf("Removed notes for %s.", capname)
      else
        meri.sv.player_notes[name] = args.leftover_full
        printf("Saved notes for %s.", capname)
      end
    else
      if meri.sv.player_notes[name] then
        printf("Notes for %s: %s", capname, meri.sv.player_notes[name])
      else
        printf("No notes for %s.", capname)
      end
    end
  end
  if args.R then
    local channels
    if args.R:sub(1, 1) == '~' then
      channels = lmeri.lfrp_channel(false, args.R:sub(2, -1))
    else
      channels = lmeri.lfrp_channel(true, args.R)
    end
    if #channels > 0 then
      printf("Channels:")
      for i = 1, #channels do
        printfs("  %s\n", channels[i])
      end
    else
      printf("No channels subscribed.")
    end
    meri.lfrp_update_channel_display()
    did_something = true
  end
  if args.S then
    while #args.leftover_args > 1 do
      local field = args.leftover_args[1]
      local value = args.leftover_args[2]
      table.remove(args.leftover_args, 2)
      table.remove(args.leftover_args, 1)
      lmeri.flaunt(field, value, meri.save_failed_callback)
    end
    did_something = true
  end
  if args.C then
    lmeri.flaunt('currently', args.leftover_full, meri.save_failed_callback)
    printf("Okay, now doing: %s", args.leftover_full)
    args.leftover_full = ""
    args.leftover_args = {}
    did_something = true
  end
  if args.B then
    did_something = true
    args.B = args.B:upper()
    if valid_point_names[args.B] then
      meri.sv.button_point = args.B
      -- and we assume you want it there exactly if you didn't say otherwise.
      if (meri.sv.button_x ~= 0 and not args.X) or
         (meri.sv.button_y ~= 0 and not args.Y) then
        printf("Since the button's origin point is changing, zeroing offset.")
      end
      meri.sv.button_x = 0
      meri.sv.button_y = 0
      replace_button = true
    else
      local valid = {}
      for k, v in pairs(valid_point_names) do
        valid[#valid + 1] = k
      end
      table.sort(valid)
      printf("Point should be one of: %s", table.concat(valid, ", "))
    end
  end
  if args.X then
    meri.sv.button_x = args.X
    did_something = true
    replace_button = true
  end
  if args.Y then
    meri.sv.button_y = args.Y
    did_something = true
    replace_button = true
  end
  if replace_button then
    meri.place_button('player.target')
  end
  if args.b then
    if meri.sv then
      meri.sv.suppress_button = not meri.sv.suppress_button
      if meri.sv.suppress_button then
        printf("Will not show the <RP> button under target portrait.")
      else
        printf("Will show the <RP> button under target portrait.")
      end
    else
      printf("Can't toggle button until saved variables are loaded, sorry.")
    end
    did_something = true
  end
  if args.t then
    if meri.sv then
      meri.sv.suppress_tooltip = not meri.sv.suppress_tooltip
      if meri.sv.suppress_tooltip then
        printf("Will not show additional tooltip data.")
      else
        printf("Will show additional tooltip data.")
      end
    else
      printf("Can't toggle tooltip until saved variables are loaded, sorry.")
    end
    did_something = true
  end
  if args.c then
    -- get data set up so we can save changes
    meri.populate_editor(meri.my_name, meri.my_local_data)
    for c in args.c:gmatch(".") do
      meri.check_box(meri.editor, 'flags', c, false, true)
    end
    meri.save_changes()
    did_something = true
  end
  if args.s then
    -- get data set up so we can save changes
    meri.populate_editor(meri.my_name, meri.my_local_data)
    for c in args.s:gmatch(".") do
      meri.check_box(meri.editor, 'flags', c, true, true)
    end
    meri.save_changes()
    did_something = true
  end
  if args.e then
    did_something = true
    lmeri.snoop(Inspect.Unit.Detail('player').name, meri.edit_callback)
    return
  end
  if args.f then
    did_something = true
    local known = lmeri.flags()
    local keys = {}
    for k, v in pairs(known) do
      keys[#keys + 1] = k
    end
    table.sort(keys)
    local output = ""
    for i = 1, #keys do
      output = output .. sprintf("%s: %s\n", keys[i], known[keys[i]].description)
    end
    printf("Known flags:\n%s", output)
  end
  if args.L or args.l then
    local known = lmeri.list(args.L)
    table.sort(known)
    for i = 1, #known do
      known[i] = known[i]:sub(1, 1):upper() .. known[i]:sub(2)
    end
    if args.L then
      message = "Users matching %s:\n%s"
    else
      message = "Known users:%.0s\n%s"
    end
    printf(message, args.L or "", table.concat(known, "\n"))
    did_something = true
  end
  if args.h then
    did_something = true
    printfh('<font color="#bb44ff">Welcome!</font>')
    printfhs("<font color=\"#bb66ff\">If you are new to roleplaying in MMOs, or even just new to Rift, you may want to familiarize yourself with the roleplaying community and its habits and expectations.</font>")
    printfhs("<font color=\"#bb66ff\">One starting point is the Role-play forums on the official forums (under the Adventures in Telara heading); that's at:</font> <font color=\"#88ccff\">http://forums.riftgame.com/game-discussions/adventures-telara/role-play-forums/</font>")
    printfhs("<font color=\"#bb66ff\">If you're new, you may want to check the Newbie flag in the Flags section of your bio, and if you are using a character from another game or setting, the Non-Lore flag will remind people that you are not trying to follow Rift's lore.</font>")
    printfhs("<font color=\"#bb66ff\">Likely commands:</font>")
    printfhs("<font color=\"#aaaaaa\">Create note:</font> meri -n &lt;player&gt; &lt;note&gt;</font>")
    printfhs("<font color=\"#aaaaaa\">Open LFRP tool:</font> meri -r</font>")
    printfhs("<font color=\"#aaaaaa\">Add LFRP channel:</font> meri -R &lt;channel&gt;</font>")
    printfhs("<font color=\"#aaaaaa\">Remove LFRP channel:</font> meri -R ~&lt;channel&gt;</font>")
    printfhs("<font color=\"#aaaaaa\">Edit data:</font> meri -e</font>")
    printfhs("<font color=\"#aaaaaa\">Show data:</font> meri &lt;player&gt;</font>")
  end
  if args.v then
    did_something = true
    printf("Merisioux version %s, LibMerisioux version %s.",
      meri.version, lmeri and lmeri.version or "unknown")
  end
  if not did_something then
    if #args.leftover_args > 1 then
      printf([[
Usage:
  Edit data: /meri -e
  View character: /meri <character>
  Clear flags: /meri -c <flags>
  Set flags: /meri -s <flags>
  List flags: /meri -f
  List known players: /meri -l
  List known players matching flags: /meri -L <flags>
  Toggle the "RP" button: /meri -b
  Toggle the extra tooltip: /meri -t
  For instance, "/meri -s o" would set the OOC flag.
]])
      printf("/meri [-e] [charname]")
      printf("       /meri [-e] [charname]")
    elseif #args.leftover_args == 1 then
      lmeri.snoop(args.leftover_args[1], meri.edit_callback)
    else
      -- cheat!  We try to grab the target.
      meri.display_data('player.target', true)
    end
  end
end

--[[
  Prefix Name Suffix
  "Title"
  Description ...
  [ ] Bio   [ ] Meta
]]--

function meri.tooltip_update(name, data, error)
  if not data or not next(data) then
    data = { error = error or "Unknown error." }
  end
  local counter = 0
  local capname = capitalize(name)

  -- add locally-stored notes, which are displayed even for people with
  -- no RP data.
  local maxwidth = 0

  local note = meri.sv.player_notes[name]
  if note and #note > 0 then
    counter = counter + 1
    meri.tooltip.labels[counter]:ClearWidth()
    if #note > 58 then
      meri.tooltip.labels[counter]:SetText('Note: ' .. string.sub(note, 1, 57) .. '...')
    else
      meri.tooltip.labels[counter]:SetText('Note: ' .. note)
    end
    meri.tooltip.labels[counter]:SetVisible(true)
    meri.tooltip.labels[counter]:SetFontColor(0.85, 0.85, 0.85)
    if meri.tooltip.labels[counter]:GetWidth() > 348 then
      meri.tooltip.labels[counter]:SetWidth(348)
    end
    maxwidth = meri.tooltip.labels[counter]:GetWidth()
  end

  if data.message then
    meri.tooltip.frame:SetBackgroundColor(0.3, 0.3, 0.3, 1.0)
    meri.tooltip.subframe:ClearWidth()
    counter = counter + 1
    meri.tooltip.labels[counter]:ClearWidth()
    meri.tooltip.labels[counter]:SetText(data.message)
    meri.tooltip.labels[counter]:SetVisible(true)
    meri.tooltip.labels[counter]:SetFontColor(0.75, 0.75, 0.65)
    if meri.tooltip.labels[counter]:GetWidth() > 348 then
      meri.tooltip.labels[counter]:SetWidth(348)
    end
    maxwidth = max(maxwidth, meri.tooltip.labels[counter]:GetWidth())
    meri.tooltip.subframe:SetWidth(maxwidth + 2)
  elseif data.error then
    meri.tooltip.frame:SetBackgroundColor(0.2, 0.2, 0.2, 1.0)
    meri.tooltip.subframe:ClearWidth()
    counter = counter + 1
    meri.tooltip.labels[counter]:ClearWidth()
    meri.tooltip.labels[counter]:SetText("(No RP data.)")
    meri.tooltip.labels[counter]:SetVisible(true)
    meri.tooltip.labels[counter]:SetFontColor(0.65, 0.65, 0.65)
    if meri.tooltip.labels[counter]:GetWidth() > 348 then
      meri.tooltip.labels[counter]:SetWidth(348)
    end
    maxwidth = max(maxwidth, meri.tooltip.labels[counter]:GetWidth())
    meri.tooltip.subframe:SetWidth(maxwidth + 2)
  else
    meri.tooltip.subframe:SetWidth(350)
    meri.tooltip.labels[1]:SetWidth(348)
    meri.tooltip.labels[2]:SetWidth(348)
    meri.tooltip.frame:SetBackgroundColor(0.4, 0.5, 0.3, 1.0)
    
    if data.flags and data.flags ~= '' then
      local has_flags = {}
      if lmeri.lfrp_match(data.flags) then
	has_flags[#has_flags + 1] = "(LFRP)"
      end
      for c in string.gmatch(meri.sorted_flags, '.') do
        if string.find(data.flags, c) then
	  has_flags[#has_flags + 1] = meri.flags[c].name
	end
      end
      counter = counter + 1
      meri.tooltip.labels[counter]:SetText(table.concat(has_flags, ', '))
      meri.tooltip.labels[counter]:SetVisible(true)
      meri.tooltip.labels[counter]:SetFontColor(0.65, 0.65, 0.85)
    end

    if data.biography or data.meta or data.cgender then
      local display = ''
      if data.biography and data.biography ~= '' then
        display = display .. '(Has Bio) '
      end
      if data.meta and data.meta ~= '' then
        display = display .. ' (Has Meta)'
      end
      if data.cgender and data.cgender ~= '' then
        display = display .. ' [' .. data.cgender .. ']'
      end
      if #display > 0 then
        counter = counter + 1
        meri.tooltip.labels[counter]:SetText(display)
        meri.tooltip.labels[counter]:SetVisible(true)
        meri.tooltip.labels[counter]:SetFontColor(0.85, 0.85, 0.65)
      end
    end

    if data.description and data.description ~= '' then
      counter = counter + 1
      if #data.description > 200 then
        meri.tooltip.labels[counter]:SetText(string.sub(data.description, 1, 197) .. '...')
      else
        meri.tooltip.labels[counter]:SetText(data.description)
      end
      meri.tooltip.labels[counter]:SetVisible(true)
      meri.tooltip.labels[counter]:SetFontColor(0.75, 0.75, 0.75)
    end

    -- add locally-stored notes
    local egg = meri.easter_eggs[name]
    if egg and #egg > 0 then
      counter = counter + 1
      meri.tooltip.labels[counter]:SetText(egg)
      meri.tooltip.labels[counter]:SetVisible(true)
      meri.tooltip.labels[counter]:SetFontColor(0.8, 0, 0)
    end

    if data.currently and data.currently ~= '' then
      counter = counter + 1
      if #data.currently > 64 then
        meri.tooltip.labels[counter]:SetText(string.sub(data.currently, 1, 61) .. '...')
      else
        meri.tooltip.labels[counter]:SetText(data.currently)
      end
      meri.tooltip.labels[counter]:SetVisible(true)
      meri.tooltip.labels[counter]:SetFontColor(0.65, 0.85, 0.65)
    end

    if data.title and data.title ~= '' then
      counter = counter + 1
      meri.tooltip.labels[counter]:SetText('"' .. data.title .. '"')
      meri.tooltip.labels[counter]:SetVisible(true)
      meri.tooltip.labels[counter]:SetFontColor(0.85, 0.85, 0.35)
    end

    counter = counter + 1
    meri.tooltip.labels[counter]:SetText(lmeri.prettyname(name, data))
    meri.tooltip.labels[counter]:SetFontColor(0.85, 0.85, 0.85)
    meri.tooltip.labels[counter]:SetVisible(true)
  end

  meri.tooltip.subframe:SetPoint("TOP", meri.tooltip.labels[counter], "TOP", nil, -1)

  if counter < #meri.tooltip.labels then
    for i = counter + 1, #meri.tooltip.labels do
      meri.tooltip.labels[i]:SetVisible(false)
    end
  end
  return true
end

-- occasionally people make me think they should have some kind of
-- recognition, so we make their tooltips shine.
meri.glow_colors = {
    merisioux = { colorR = .6, colorG = 0, colorB = 1 },
    iskre = { colorR = 1, colorG = 0.4, colorB = 0 },
}

meri.easter_eggs = {
    morrandir = "Voiced by: Vincent Price"
}

meri.glow = {
	colorR = .6, colorG = 0, colorB = 1,
	blurX = 0.2, blurY = 0.2,
	strength = 0.2
}
local glow_dir = 1

function meri.show_tooltip(name, data, error)
  if not meri.tooltip then
    meri.tooltip = {}
    meri.tooltip.labels = {}
    meri.tooltip.ctx = UI.CreateContext("Merisioux Tooltip")
    if not meri.tooltip.ctx then
      printf("Couldn't create a tooltip.  Oops.")
      meri.tooltip = nil
      return
    end
    meri.tooltip.ctx:SetStrata('tutorial')
    meri.tooltip.frame = UI.CreateFrame('Frame', 'Tooltip', meri.tooltip.ctx)
    meri.tooltip.frame:SetBackgroundColor(0.4, 0.4, 0.3, 1.0)
    meri.tooltip.subframe = UI.CreateFrame('Frame', 'Tooltip Subframe', meri.tooltip.frame)
    meri.tooltip.subframe:SetBackgroundColor(0.1, 0.1, 0.1, 1.0)
    meri.tooltip.subframe:SetWidth(350)

    meri.tooltip.frame:SetPoint("TOPLEFT", meri.tooltip.subframe, "TOPLEFT", -2, -2)
    meri.tooltip.frame:SetPoint("BOTTOMRIGHT", UI.Native.Tooltip, "TOPRIGHT", -5, 5)
    meri.tooltip.subframe:SetPoint("BOTTOMRIGHT", meri.tooltip.frame, "BOTTOMRIGHT", -2, -2)

    table.insert(meri.tooltip.labels, UI.CreateFrame('Text', 'Label 1', meri.tooltip.subframe))
    meri.tooltip.labels[1]:SetPoint("BOTTOMCENTER", meri.tooltip.subframe, "BOTTOMCENTER", 0, -2)
    meri.tooltip.labels[1]:SetFontSize(meri.tooltip.labels[1]:GetFontSize() + 2)
    meri.tooltip.labels[1]:SetFontColor(0.85, 0.85, 0.85)
    meri.tooltip.labels[1]:SetWordwrap(true)
    meri.tooltip.labels[1]:SetText(name or "Unknown")
    meri.tooltip.labels[1]:SetWidth(348)
    for counter = 2, 8 do
      table.insert(meri.tooltip.labels, UI.CreateFrame('Text', 'Label ' .. counter, meri.tooltip.subframe))
      meri.tooltip.labels[counter]:SetPoint("BOTTOMCENTER", meri.tooltip.labels[counter - 1], "TOPCENTER", 0, 1)
      meri.tooltip.labels[counter]:SetFontSize(meri.tooltip.labels[counter]:GetFontSize() + 2)
      meri.tooltip.labels[counter]:SetWordwrap(true)
      meri.tooltip.labels[counter]:SetFontColor(0.85, 0.85, 0.85)
      meri.tooltip.labels[counter]:SetWidth(348)
    end
  end
  if meri.glow_colors[name] then
    local color = meri.glow_colors[name]
    meri.glow.colorR = color.colorR
    meri.glow.colorG = color.colorG
    meri.glow.colorB = color.colorB
    Command.Event.Detach(Event.System.Update.Begin, meri.shimmer)
    Command.Event.Attach(Event.System.Update.Begin, meri.shimmer, "meri_shimmer")
    for i = 1, #meri.tooltip.labels do
      meri.tooltip.labels[i]:SetEffectGlow(meri.glow)
    end
  else
    Command.Event.Detach(Event.System.Update.Begin, meri.shimmer)
    for i = 1, #meri.tooltip.labels do
      meri.tooltip.labels[i]:SetEffectGlow(nil)
    end
  end
  meri.tooltip_update(name, data, error)
  meri.tooltip.frame:SetVisible(meri.tooltip_is_visible and not (meri.sv and meri.sv.suppress_tooltip))
end

function meri.shimmer()
  meri.glow.strength = meri.glow.strength + glow_dir * 0.02
  meri.glow.blurX = meri.glow.blurX + glow_dir * 0.05
  meri.glow.blurY = meri.glow.blurY + glow_dir * 0.05
  if meri.glow.strength >= 0.8 then
    glow_dir = -1
  elseif meri.glow.strength < 0 then
    meri.glow.strength = 0
    meri.glow.blurX = 0
    meri.glow.blurY = 0
    glow_dir = 1
  end
  for i = 1, #meri.tooltip.labels do
    meri.tooltip.labels[i]:SetEffectGlow(meri.glow)
  end
end

function meri.tooltip_visible(flag)
  flag = flag and not (meri.sv and meri.sv.suppress_tooltip)
  meri.tooltip_is_visible = flag
  if not flag then
    Command.Event.Detach(Event.System.Update.Begin, meri.shimmer)
  end
  if meri.tooltip then
    meri.tooltip.frame:SetVisible(flag)
  end
end

function meri.snoop_callback(char, data, error)
  if data then
    printf("Got data for %s:", char or 'nil')
  else
    printf("Error trying to get data for %s: %s", char or 'nil', error or 'nil')
  end
end

function meri.display_data(unit, force)
  if not unit then
    return
  end
  local details = Inspect.Unit.Detail(unit)
  if details and details.player then
    lmeri.snoop(details.name, meri.edit_callback)
    return
  end
  if force then
    printf("Need a player target.")
  end
end

meri.targeting = nil

function meri.target_changed(event, unit)
  -- it will come back shortly if appropriate
  meri.unitbuttons.buttons['player.target']:SetVisible(false)
  if not unit then
    meri.targeting = nil
    return
  end
  local details = Inspect.Unit.Detail(unit)
  if details and details.player then
    -- record who we think we're investigating, so we don't show the
    -- wrong button
    meri.targeting = strlower(details.name)
    if meri.editor and meri.editor.window:GetVisible() then
      meri.display_data(unit)
    end
    if meri.unitbuttons.buttons['player.target'] then
      lmeri.snoop(details.name, meri.unitbuttons.buttons['player.target'].callback)
    end
  end
end

function meri.tooltip_handler(event, tt_type, shown, buff)
  if tt_type == 'unit' then
    target = Inspect.Unit.Detail(shown)
    if target and target.player then
      meri.show_tooltip(target.name, { message = "Loading..." }, nil)
      meri.tooltip_visible(true)
      lmeri.snoop(target.name, meri.show_tooltip)
    else
      meri.tooltip_visible(false)
    end
  else
    if Inspect.Time.Frame() ~= meri.this_frame then
      meri.tooltip_visible(false)
    end
  end
end

local unit_to_portrait = {
  ['player.target'] = UI.Native.PortraitTarget,
}

function meri.place_button(unit)
  if unit_to_portrait[unit] and meri.unitbuttons and meri.unitbuttons.buttons and meri.unitbuttons.buttons[unit] then
    local sv = meri.sv or { }
    meri.unitbuttons.buttons[unit]:SetPoint("CENTER", unit_to_portrait[unit], sv.button_point or "CENTER", sv.button_x or 72, sv.button_y or 60)
  end
end

meri.unitbuttons = {}
meri.unitbuttons.ctx = UI.CreateContext('Merisioux Unit Buttons')
meri.unitbuttons.ctx:SetStrata('notify')
meri.unitbuttons.buttons = {}
meri.unitbuttons.buttons['player.target'] = UI.CreateFrame('RiftButton', 'Merisioux', meri.unitbuttons.ctx)
meri.unitbuttons.buttons['player.target']:SetWidth(50)
local rp_text = UI.CreateFrame('Text', 'Merisioux', meri.unitbuttons.buttons['player.target'])
rp_text:SetPoint("CENTER", meri.unitbuttons.buttons['player.target'], "CENTER")
rp_text:SetFontSize(16)
rp_text:SetFontColor(1, .9, .7, 1)
rp_text:SetText("RP")
rp_text:SetEffectGlow({ strength = 2, blurX = 1, blurY = 1 })
meri.unitbuttons.buttons['player.target']:SetText('')
meri.unitbuttons.buttons['player.target']:SetVisible(false)
meri.unitbuttons.buttons['player.target'].callback = function(c, d, e) meri.button_callback('player.target', c, d, e) end
meri.unitbuttons.buttons['player.target']:EventAttach(Event.UI.Input.Mouse.Left.Click, function() meri.display_data('player.target') end, "display target")
meri.place_button('player.target')

Library.LibGetOpt.makeslash("bB:c:CefhlL:n:rR:s:StvX#Y#", "Merisioux", "meri", meri.slashcommand)

Command.Event.Attach(Event.Tooltip, meri.tooltip_handler, "tooltip hook")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, meri.variables_loaded, "variable loaded hook")
local change = Library.LibUnitChange.Register('player.target')
if change then
  Command.Event.Attach(Event.LibUnitChange.player.target.Change, meri.target_changed, "target change handler")
end
